#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/unvmath.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "swterm.h"


swt_bridge_struct *BridgeNew(void);
void BridgeDraw(swt_bridge_struct *bw, int portion);
void BridgeDrawSchedualed(
	swt_bridge_struct *bw,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int BridgeManage(swt_bridge_struct *bw, CEvent *event);
void BridgeSetFocus(swt_bridge_struct *bw, int focus);
void BridgeMap(swt_bridge_struct *bw);
void BridgeUnmap(swt_bridge_struct *bw);
void BridgeDelete(swt_bridge_struct *bw);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new bridge.
 */
swt_bridge_struct *BridgeNew(void)
{
	const int	total_messages = 100,
			msg_lines_visible = 5;
	int i;
	CDisplay *display = NULL;
	Window parent;
	Prompt *prompt;
	swt_msg_struct *mesg;
	swt_bridge_struct *bw = SWT_BRIDGE(calloc(
	    1, sizeof(swt_bridge_struct)
	));
	if(bw == NULL)
	    return(NULL);

	if(total_curses_displays > 0)
	    display = curses_display[0];
	if((display == NULL) || (bw == NULL))
	    return(bw);

	parent = display->root_win;

	bw->x = 0;
	bw->y = 0;
	bw->width = CWRootWidth(display);
	bw->height = CWRootHeight(display);
	bw->vs_zoom = 0.30f;
	bw->screen_type = BRIDGE_SCREEN_TYPE_MAIN_MENU;

	/* Initialize prompt */
	prompt = &bw->prompt;
	PromptInit(
	    display, prompt, parent,
	    1,
	    (int)CWRootHeight(display) - 1 - msg_lines_visible - 1,
	    (unsigned int)MAX(CWRootWidth(display) - 2, 10),
	    1,
	    256,
	    "Prompt:",
	    BridgePromptCB, NULL
	);
	bw->prompt_mode = SWT_PROMPT_MODE_NONE;

	/* Allocate message buffers */
	bw->total_messages = total_messages;
	bw->message = (swt_msg_struct **)calloc(
	    bw->total_messages,
	    sizeof(swt_msg_struct *)
	);
	if(bw->message == NULL)
	{
	    bw->total_messages = 0;
	    return(bw);
	}

	for(i = 0; i < bw->total_messages; i++)
	{
	    bw->message[i] = mesg = (swt_msg_struct *)calloc(
		1, sizeof(swt_msg_struct)
	    );
	    if(mesg == NULL)
		break;

	    mesg->message = NULL;
	    mesg->color_num = xsw_color.messages;
	}

	bw->msg_lines_visible = msg_lines_visible;
	bw->message_scroll_pos = bw->total_messages - msg_lines_visible;


	return(bw);
}

/*
 *	Draws the bridge window.
 */
void BridgeDraw(swt_bridge_struct *bw, int portion)
{
	CDisplay *display = NULL;

	if(bw == NULL)
	    return;

	if(total_curses_displays > 0)
	    display = curses_display[0];
	if((display == NULL) || (bw == NULL))
	    return;

	/* Check which screen to draw */
	switch(bw->screen_type)
	{
	  default:
	    /* Redraw messages */
	    BridgeDrawMessages(display, bw);

	    /* Redraw title */
	    BridgeDrawTitle(display, bw);

	    /* Redraw everything else */
	    BridgeDrawSchedualed(
		bw,
		cur_millitime, cur_systime, time_compensation
	    );

	    break;
	}
}

/*
 *	Redraws things on the bridge window that need to be
 *	redrawn often.
 */
void BridgeDrawSchedualed(
	swt_bridge_struct *bw,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num;
	xsw_object_struct *obj;
	CDisplay *display = NULL;
	xsw_net_parms_struct *net = &net_parms;

	if(bw == NULL)
	    return;

	if(total_curses_displays > 0)
	    display = curses_display[0];
	if(display == NULL)
	    return;

	/* Change screen type as needed */
	switch(bw->screen_type)
	{
	  case BRIDGE_SCREEN_TYPE_MAIN_MENU:
	    if(net->connection_state != XSW_CON_STATE_NOT_CONNECTED)
		bw->screen_type = BRIDGE_SCREEN_TYPE_VIEWSCREEN;
	    break;

	  case BRIDGE_SCREEN_TYPE_VIEWSCREEN:
	    if(net->connection_state == XSW_CON_STATE_NOT_CONNECTED)
		bw->screen_type = BRIDGE_SCREEN_TYPE_MAIN_MENU;
	    break;
	}

	/* Check which screen to draw */
	switch(bw->screen_type)
	{
	  case BRIDGE_SCREEN_TYPE_MAIN_MENU:
	    BridgeDrawClearVS(display, bw);
	    BridgeDrawStatusBar(display, bw);
	    break;

	  case BRIDGE_SCREEN_TYPE_VIEWSCREEN:
	    obj_num = net->player_obj_num;
	    obj = DBIsObjectGarbage(obj_num) ?
		NULL : xsw_object[obj_num];

	    /* Clear viewscreen area */
	    BridgeDrawClearVS(display, bw);

	    /* Draw player stats */
	    BridgeDrawPlayerStats(display, bw, obj_num);

	    /* Draw locked on subject object */
	    BridgeDrawSubjectStats(
	        display,
		bw,
	        (obj != NULL) ? obj->locked_on : -1
	    );

	    /* Draw objects on viewscreen */
	    BridgeDrawVSObjects(display, bw, cur_ms, cur_sec, tc);

	    /* Draw status bar */
	    BridgeDrawStatusBar(display, bw);
	    break;

	  case BRIDGE_SCREEN_TYPE_HELP:
	    BridgeDrawClearVS(display, bw);
	    BridgeDrawHelp(display, bw);
	    BridgeDrawStatusBar(display, bw);
	    break;
	}
}

/*
 *	Manages the bridge window.
 */
int BridgeManage(swt_bridge_struct *bw, CEvent *event)
{
	int events_handled = 0;
	int key;
	int obj_num;
	xsw_object_struct *obj;
	CDisplay *display = NULL;
	xsw_gctl_struct *gctl = xsw_gctl;
	xsw_net_parms_struct *net = &net_parms;


	if(total_curses_displays > 0)
	    display = curses_display[0];
	if((display == NULL) || (event == NULL) || (bw == NULL))
	    return(events_handled);

	/* Not mapped or not in focus? */
	if(!bw->map_state || !bw->has_focus)
	    return(events_handled);

/* Debug, check if key is supported */
if(option.debug_kb && (event->type == CKeyPress))
{
 char s[80];
 key = event->ckey.key;
 sprintf(s, "Debug key 0x%.8x", (u_int32_t)key);
 XSWPrint(s);
}

	/* Manage the prompt first */
	if(bw->prompt_mode != SWT_PROMPT_MODE_NONE)
	{
	    key = event->ckey.key;

	    /* Cancel? */
	    if((event->type == CKeyPress) &&
	       (key == KEY_ESC)
	    )
	    {
		PromptUnmap(display, &bw->prompt);
		bw->prompt_mode = SWT_PROMPT_MODE_NONE;
	    }
	    else
	    {
	        events_handled += PromptManage(
		    display, &bw->prompt, event
	        );
	    }
	    if(events_handled > 0)
		return(events_handled);
	}


	/* Manage event normally */
	switch(event->type)
	{
	  case CKeyPress:
	    key = event->ckey.key;

	    /* Help screen */
	    if((key == KEY_F(1)) || (key == '?') || (key == 0x00000113))
	    {
		if(bw->screen_type == BRIDGE_SCREEN_TYPE_HELP)
		    bw->screen_type = BRIDGE_SCREEN_TYPE_VIEWSCREEN;
		else
		    bw->screen_type = BRIDGE_SCREEN_TYPE_HELP;
	    }
	    /* Turn */
	    else if(key == KEY_RIGHT)
	    {
#if 0
		if(gctl != NULL)
		{
		    gctl->turn += 0.1f;
		    if(gctl->turn > 1.0f)
			gctl->turn = 1.0f;
		}
#endif
/* Turn explicitly */
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;

		if(obj != NULL)
		{
		    obj->heading = SANITIZERADIANS(
			obj->heading + (0.043 *
			    time_compensation)
		    );
		}
	    }
	    else if(key == KEY_LEFT)
	    {
#if 0
		if(gctl != NULL)
		{
		    gctl->turn -= 0.1f;
		    if(gctl->turn < -1.0f)
			gctl->turn = -1.0f;
		}
#endif
/* Turn explicitly */
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
	     
		if(obj != NULL)
		{
		    obj->heading = SANITIZERADIANS(
			obj->heading - (0.043 *
			    time_compensation)
		    );
		}
	    }
	    /* Throttle */
	    else if(key == KEY_UP)
	    {
		if(gctl != NULL)
		{
		    gctl->throttle += 0.1f;
		    if(gctl->throttle > 1.0f)
			gctl->throttle = 1.0f;
		}
	    }
	    else if(key == KEY_DOWN)
	    {
		if(gctl != NULL)
		{
		    gctl->throttle -= 0.1f;
		    if(gctl->throttle < -1.0f)
			gctl->throttle = -1.0f;
		}
	    }
	    /* Retro thrusters */
	    else if(key == 'b')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    obj->velocity = 0.0f;
		    REngObjectStopped(
			obj, obj,
			cur_millitime, cur_systime, time_compensation
		    );
		}
	    }
	    /* Viewscreen zoom */
	    else if(key == '-')
	    {
		bw->vs_zoom += 0.01f;
		if(bw->vs_zoom > VS_ZOOM_MAX)
		    bw->vs_zoom = VS_ZOOM_MAX;
	    }
	    else if(key == '_')
	    {
		bw->vs_zoom += 0.10f;
		if(bw->vs_zoom > VS_ZOOM_MAX)
		    bw->vs_zoom = VS_ZOOM_MAX;
	    }
	    else if(key == '=')
	    {
		bw->vs_zoom -= 0.01;
		if(bw->vs_zoom < VS_ZOOM_MIN)
		    bw->vs_zoom = VS_ZOOM_MIN;
	    }
	    else if(key == '+')
	    {
		bw->vs_zoom -= 0.10f;
		if(bw->vs_zoom < VS_ZOOM_MIN)
		    bw->vs_zoom = VS_ZOOM_MIN;
	    }
	    /* Select weapon 1 */
	    else if(key == '1')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 0);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 2 */
	    else if(key == '2')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 1);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 3 */
	    else if(key == '3')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 2);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 4 */
	    else if(key == '4')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 3);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 5 */
	    else if(key == '5')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 4);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 6 */
	    else if(key == '6')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 5);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 7 */
	    else if(key == '7')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 6);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 8 */
	    else if(key == '8')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 7);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 9 */
	    else if(key == '9')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 8);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select weapon 10 */
	    else if(key == '0')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(obj_num, 9);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select previous weapon */
	    else if(key == ',')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(
			obj_num,
			((obj->selected_weapon > 0) ?
			    obj->selected_weapon - 1 :
			    obj->total_weapons - 1
			)
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Select next weapon */
	    else if(key == '.')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSelectWeapon(
			obj_num,
			(((obj->selected_weapon + 1) >= obj->total_weapons) ?
			    0 :
			    obj->selected_weapon + 1
			)
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Fire weapon */
	    else if(key == ' ')
	    {
		/* This section of code comes From REngAction()
		 * we cannot call it since the xsw_gctl for fire weapon
		 * cannot be set to 1 and held there due to key board
		 * control limitations.
		 */
		int sel_wep_num;
		time_t dt;
		xsw_weapon_struct *wep_ptr;

		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;

		while(obj != NULL)
		{
		    /* Get pointer to selected weapon */
		    sel_wep_num = obj->selected_weapon;
		    if((sel_wep_num < 0) ||
		       (sel_wep_num >= obj->total_weapons)
		    )
		    break;
 
		    wep_ptr = obj->weapons[sel_wep_num];
		    if(wep_ptr == NULL)
			break;

		    /* Check if weapons are online */
		    if(!player_values.weapons_online)
		    {
			/* Weapons are offline, play error sound */
			dt = wep_ptr->last_used +
			    MIN(wep_ptr->delay, 1000l);
			if(dt <= cur_millitime)
			{
			    BridgeWarnWeaponsOffline(bw, obj_num);
			    wep_ptr->last_used = cur_millitime;
			}
			break;
		    }

		    /* Check if we've waited the weapon's delay period */
		    dt = wep_ptr->last_used + wep_ptr->delay;
		    if(dt >= cur_millitime)
			break;

		    /* Send the fire weapon command to server */
		    NetSendFireWeapon(
			obj_num,
			player_values.weapon_freq
		    );

		    /* Mark last millitime this weapon was used.
		     * The last used time sent by the server is ignored and
		     * the client keeps track of the last used time.  Since
		     * the firing times and weapon delay times are synced
		     * (or atleast reasonably synced) there shouldn't be any
		     * problem.
		     */
		    wep_ptr->last_used = cur_millitime;


		    break;
		}
	    }
	    /* Message box scroll */
	    else if(key == KEY_PAGEUP)
	    {
		bw->message_scroll_pos--;
		if(bw->message_scroll_pos < 0)
		    bw->message_scroll_pos = 0;
		BridgeDrawMessages(display, bw);
	    }
	    else if(key == KEY_PAGEDOWN)
	    {
		const int	msg_lines_visible = bw->msg_lines_visible,
				last_line = bw->total_messages - msg_lines_visible;
		bw->message_scroll_pos++;
		if(bw->message_scroll_pos > last_line)
		    bw->message_scroll_pos = last_line;
		BridgeDrawMessages(display, bw);
	    }
	    else if((key == KEY_END) || (key == 0x00000005))
	    {
		const int msg_lines_visible = bw->msg_lines_visible;
		bw->message_scroll_pos = bw->total_messages - msg_lines_visible;
		BridgeDrawMessages(display, bw);
	    }
	    /* Exit */
	    else if((key == 'x') || (key == 0x00000003))
	    {
		BridgeMapPrompt(bw, SWT_PROMPT_MODE_EXIT);
	    }
	    /* Connect */
	    else if(key == KEY_F(9))
	    {
		BridgeMapPrompt(bw, SWT_PROMPT_MODE_CONNECT);
	    }
	    /* Disconnect */
	    else if(key == KEY_F(10))
	    {
		if(!XSW_CON_STATE_IS_NOT_CONNECTED(net))
		{
		    net->disconnect_send_count++;
		    if(net->disconnect_send_count >= 5)
		    {
			SWTDisconnect();
		    }
		    else
		    {
		        if(net->disconnect_send_count == 1)
			{
			    char *s = (char *)malloc(
				STRLEN(net->address) + 80
			    );
			    sprintf(
				s,
"Disconnecting from %s...",
				net->address
			    );
			    XSWPrint(s);
			    free(s);
			}
			NetSendLogout();
		    }
		}
		else
		{
		    XSWPrintWarning(
"Not connected."
		    );
		}
	    }
	    /* Refresh */
	    else if(key == KEY_F(11))
	    {
		SWTRefresh(); 
	    }
	    /* Connect last */
	    else if(key == KEY_F(12))
	    {
		if(net->connection_state == XSW_CON_STATE_NOT_CONNECTED)
		    NetConnect(net->address, net->port, 45l);
	    }
	    /* Auto interval state toggle */
	    else if(key == 'a')
	    {
		auto_interval_tune.state = !auto_interval_tune.state;
	    }
	    /* Network stream interval tunning */
	    else if(key == '[')
	    {
		net->net_int -= 25;
		if(net->net_int < XSW_SWSERV_MIN_UPDATE_INT)
		    net->net_int = XSW_SWSERV_MIN_UPDATE_INT;
		NetSendSetInterval();
	    }
	    /* Network stream interval tunning */
	    else if(key == '{')
	    {
		net->net_int -= 100;
		if(net->net_int < XSW_SWSERV_MIN_UPDATE_INT)
		    net->net_int = XSW_SWSERV_MIN_UPDATE_INT;

		NetSendSetInterval();
	    }
	    /* Network stream interval tunning */
	    else if(key == ']')
	    {
		net->net_int += 25;
		if(net->net_int > XSW_SWSERV_MAX_UPDATE_INT)
		    net->net_int = XSW_SWSERV_MAX_UPDATE_INT;
		NetSendSetInterval();
	    }
	    /* Network stream interval tunning */
	    else if(key == '}')
	    {
		net->net_int += 100;
		if(net->net_int > XSW_SWSERV_MAX_UPDATE_INT)
		    net->net_int = XSW_SWSERV_MAX_UPDATE_INT;
		NetSendSetInterval();
	    }
	    /* Lock nearest */
	    else if((key == 'n') || (key == 'N'))
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		while(obj != NULL)
		{
		    int tar_obj_num;

#define MATCH_NEAREST_TYPE(_obj_num_,_type_)	\
 UNVMatchObjectNearest(				\
  xsw_object, total_objects,			\
  (_obj_num_), (_type_)				\
 )

		    /* If the shift key is not held then find the
		     * nearest controlled or player object, projectile
		     * weapon object, dynamic object, or static object
		     */
		    if(key == 'n')
		    {
			/* Find the nearest controlled or player
			 * object
			 */
			tar_obj_num = MATCH_NEAREST_TYPE(
			    obj_num, SW_OBJ_TYPE_CONTROLLED
			);
			if(tar_obj_num < 0)
			    tar_obj_num = MATCH_NEAREST_TYPE(
				obj_num, SW_OBJ_TYPE_PLAYER
			    );
			if(!DBIsObjectGarbage(tar_obj_num))
			{
			    NetSendWeaponsLock(obj_num, tar_obj_num);
			    break;
			}

			/* Find the nearest projectile weapon object */
			tar_obj_num = MATCH_NEAREST_TYPE(
			    obj_num, SW_OBJ_TYPE_WEAPON
			);
			if(!DBIsObjectGarbage(tar_obj_num))
			{
			    NetSendWeaponsLock(obj_num, tar_obj_num);
			    break;
			}

			/* Find the nearest dynamic object */
			tar_obj_num = MATCH_NEAREST_TYPE(
			    obj_num, SW_OBJ_TYPE_DYNAMIC
			);
			if(!DBIsObjectGarbage(tar_obj_num))
			{
			    NetSendWeaponsLock(obj_num, tar_obj_num);
			    break;
			}

			/* Find the nearest static object */
			tar_obj_num = MATCH_NEAREST_TYPE(
			    obj_num, SW_OBJ_TYPE_STATIC
			);
			if(!DBIsObjectGarbage(tar_obj_num))
			{
			    NetSendWeaponsLock(obj_num, tar_obj_num);
			    break;
			}
		    }

		    /* Find the nearest home object */
		    tar_obj_num = MATCH_NEAREST_TYPE(
			obj_num, SW_OBJ_TYPE_HOME
		    );
		    if(!DBIsObjectGarbage(tar_obj_num))
		    {
			NetSendWeaponsLock(obj_num, tar_obj_num);
			break;
		    }

		    /* Find the nearest wormhole object */
		    tar_obj_num = MATCH_NEAREST_TYPE(
			obj_num, SW_OBJ_TYPE_WORMHOLE
		    );
		    if(!DBIsObjectGarbage(tar_obj_num))
		    {
			NetSendWeaponsLock(obj_num, tar_obj_num);
			break;
		    }

		    /* Find the nearest link object */
		    tar_obj_num = MATCH_NEAREST_TYPE(
			obj_num, SW_OBJ_TYPE_ELINK
		    );
		    if(!DBIsObjectGarbage(tar_obj_num))
		    {
			NetSendWeaponsLock(obj_num, tar_obj_num);
			break;
		    }

		    /* Find the nearest area object */
		    tar_obj_num = MATCH_NEAREST_TYPE(
			obj_num, SW_OBJ_TYPE_AREA
		    );
		    if(!DBIsObjectGarbage(tar_obj_num))
		    {
			NetSendWeaponsLock(obj_num, tar_obj_num);
			break;
		    }

		    break;
#undef MATCH_NEAREST_TYPE
		}
		if(obj == NULL)
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Lock next */
	    else if(key == KEY_TAB)
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendWeaponsLock(obj_num, -2);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Unlock weapons */
	    else if(key == 'u')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendWeaponsUnlock(obj_num);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Lock tractor beam */
	    else if(key == 't')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendTractorBeamLock(
			obj_num,
			obj->locked_on
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Unlock tractor beam */
	    else if(key == 'U')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendTractorBeamLock(obj_num, -1);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Vector lights */
	    else if(key == 'l')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;

		if(obj != NULL)
		{
		    if(obj->lighting & SW_OBJ_LIGHT_VECTOR)
		    {
			obj->lighting &= ~SW_OBJ_LIGHT_VECTOR;
			XSWPrint("Vector lights: Off");
		    }
		    else
		    {
			obj->lighting |= SW_OBJ_LIGHT_VECTOR;
			XSWPrint("Vector lights: On");
		    }

		    NetSendSetLighting(
			obj_num,
			obj->lighting
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Strobe lights */
	    else if(key == 'p')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		 
		if(obj != NULL)
		{
		    if(obj->lighting & SW_OBJ_LIGHT_STROBE)
		    {
			obj->lighting &= ~SW_OBJ_LIGHT_STROBE;
			XSWPrint("Strobes: Off");
		    }
		    else
		    {
			obj->lighting |= SW_OBJ_LIGHT_STROBE;
			XSWPrint("Strobes: On");
		    }

		    NetSendSetLighting(
			obj_num,
			obj->lighting
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Engine state */
	    else if(key == '`')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    if(obj->engine_state > SW_ENGINE_STATE_OFF)
		    {
			NetSendSetEngine(obj_num, SW_ENGINE_STATE_OFF);
		    }
		    else if(obj->engine_state == SW_ENGINE_STATE_OFF)
		    {
			NetSendSetEngine(obj_num, SW_ENGINE_STATE_STARTING);
		    }
		    else
		    {
			XSWPrintWarning(
"Your vessel does not have engines."
			);
		    }
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Shields */
	    else if(key == 's')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;

		if(obj != NULL)
		{
		    switch(obj->shield_state)
		    {
		      case SW_SHIELD_STATE_DOWN:
			NetSendSetShields(
			    obj_num,
			    SW_SHIELD_STATE_UP,
			    obj->shield_frequency
			);
			break;
		      case SW_SHIELD_STATE_NONE:
			XSWPrintWarning(
"Your vessel does not have shields."
			);
			break;
		      default:
			NetSendSetShields(
			    obj_num,
			    SW_SHIELD_STATE_DOWN,
			    obj->shield_frequency
			);
			break;
		    }
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Cloak */
	    else if(key == 'z')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    switch(obj->cloak_state)
		    {
		      case SW_CLOAK_STATE_UP:
			NetSendSetCloak(obj_num, SW_CLOAK_STATE_DOWN);
			break;
		      case SW_CLOAK_STATE_NONE:
			XSWPrintWarning(
"Your vessel does not have a cloak."
			);
			break;
		      default:
			NetSendSetCloak(obj_num, SW_CLOAK_STATE_UP);
			break;
		    }
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Damage Control */
	    else if(key == 'c')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendSetDmgCtl(
			obj_num,
			(obj->damage_control == SW_DMGCTL_STATE_ON) ?
			    SW_DMGCTL_STATE_OFF : SW_DMGCTL_STATE_ON
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Auto Destruct */
	    else if(key == 'd')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendAutoDestruct(
			obj_num,
			(obj->auto_destruct_next > 0l) ?
			    0l : player_values.auto_destruct_start_int
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Weapons online/offline toggle */
	    else if(key == 'w')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    if(player_values.weapons_online)
		    {
			player_values.weapons_online = 0;
			XSWPrint(
			    "Weapons have been taken offline."
			);
		    }
		    else  
		    {
			player_values.weapons_online = 1;
			XSWPrint(
			    "Weapons are now online."
			);
		    }
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Send hail */
	    else if(key == 'h')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendHail(
			obj_num,
			(obj->locked_on > -1) ? obj->locked_on : -1,
			obj->com_channel
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Send general hail */
	    else if(key == 'H')
	    {
		obj_num = net->player_obj_num;
		obj = net->player_obj_ptr;
		if(obj != NULL)
		{
		    NetSendHail(
			obj_num,
			-1,
			obj->com_channel
		    );
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Set intercept */
	    else if(key == 'i')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_INTERCEPT);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Shield frequency */
	    else if(key == 'f')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_SHIELD_FREQ);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Weapon frequency */
	    else if(key == 'y')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_WEAPON_FREQ);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Set com channel */
	    else if(key == 'k')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_CHANNEL);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		   );
		}
	    }
	    /* Send message */
	    else if(key == '\'')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_MESSAGE);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
	    /* Server command */
	    else if(key == 'e')
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeMapPrompt(bw, SWT_PROMPT_MODE_SERVER_CMD);
		}
		else
		{
		    XSWPrintWarning(
"Not connected or not logged in."
		    );
		}
	    }
#if 0
	    else
	    {
		XSWPrintWarning("Unrecognized key.");
	    }
#endif

	    events_handled++;
	    break;
	}

	return(events_handled);
}

/*
 *	Changes focus of the bridge window.
 */
void BridgeSetFocus(swt_bridge_struct *bw, int focus)
{
	if(bw == NULL)
	    return;

	bw->has_focus = focus;
}

/*
 *	Maps the bridge window.
 */
void BridgeMap(swt_bridge_struct *bw)
{
	if(bw == NULL)
	    return;

	bw->map_state = 1;
	BridgeDraw(bw, 0);
}

/*
 *	Unmaps the bridge window.
 */
void BridgeUnmap(swt_bridge_struct *bw)
{
	if(bw == NULL)
	    return;

	bw->map_state = False;
}

/*
 *	Deletes the bridge.
 */
void BridgeDelete(swt_bridge_struct *bw)
{
	int i;
	CDisplay *display = NULL;
	swt_msg_struct *mesg;

	if(total_curses_displays > 0)
	    display = curses_display[0];
	if((display == NULL) || (bw == NULL))
	    return;

	PromptDestroy(display, &bw->prompt);
	bw->prompt_mode = SWT_PROMPT_MODE_NONE;

	for(i = 0; i < bw->total_messages; i++)
	{
	    mesg = bw->message[i];
	    if(mesg == NULL)
		continue;

	    free(mesg->message);
	    free(mesg);
	}
	free(bw->message);
	bw->message = NULL;
	bw->total_messages = 0;
}
