#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "cursesw.h"


CDisplay *CWInit(int argc, char **argv);
void CWShutdown(CDisplay *display);

Window CWCreateWindow(
	CDisplay *display, Window parent,
	int x, int y,
	unsigned int width, unsigned int height
);

unsigned int CWRootWidth(CDisplay *display);
unsigned int CWRootHeight(CDisplay *display);

int CWMatchColorCoeff(
	CDisplay *display,   
	const float fg_red, const float fg_green, const float fg_blue,
	const float bg_red, const float bg_green, const float bg_blue
);

void CWSetColor(
	CDisplay *display, Window w,
	const int color_num, const unsigned int flags
);
void CWMoveCursor(
	CDisplay *display, Window w,
	const int x, const int y
);
void CWDrawCharacter(
	CDisplay *display, Window w,
	const int x, const int y,
	const char c
);
void CWDrawString(
	CDisplay *display, Window w,
	const char *s
);
void CWDrawStringPosition(
	CDisplay *display, Window w,
	const int x, const int y,
	const char *s
);
void CWDrawStringPositionLimited(
	CDisplay *display, Window w,
	const int x, const int y,
	const char *s,
	const int limit
);
void CWClearArea(
	CDisplay *display, Window w,
	const int x, const int y,
	const unsigned int width, const unsigned int height
);

int CGetNextEvent(CDisplay *display, CEvent *event);

void CWSync(CDisplay *display, const unsigned int options);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Initializes the curses library, returns pointer to
 *	curses display structure or NULL on failure.
 */
CDisplay *CWInit(int argc, char **argv)
{
	int i;
	Bool no_colors = False;
	const char *arg;
	CDisplay *display;
	WINDOW *w;
/*	SCREEN *screen; */

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    if(!strcasecmp(arg, "--no-colors") ||
	       !strcasecmp(arg, "--no_colors") ||
	       !strcasecmp(arg, "--no-color") ||
	       !strcasecmp(arg, "--no_color") ||
	       !strcasecmp(arg, "-no-colors") ||
	       !strcasecmp(arg, "-no_colors") ||
	       !strcasecmp(arg, "-no-color") ||
	       !strcasecmp(arg, "-no_color")
	    )
	    {
		no_colors = True;
	    }
	}

	/* Allocate display structure */
	display = (CDisplay *)calloc(1, sizeof(CDisplay));
	if(display == NULL)
	    return(NULL);

	/* Initialize curses library */
	w = initscr();
	if(w == NULL)
	{
	    free(display);
	    return(NULL);
	}

	display->flags = 0;

	/* Initialize colors if available */
	if(has_colors() && !no_colors)
	{
	    start_color();
	    display->flags |= CFlagHasColors;
	    use_default_colors();		/* So that we can have colors
						 * in which the background is
						 * not cleared (transparent
						 * background) */
	    display->flags |= CFlagTransparentBackground;
	}

	/* Check if we can create or set our own colors */
	if(can_change_color() && !no_colors)
	    display->flags |= CFlagCanCreateColors;

	/* Create the colormap if there is color support */
	if(display->flags & CFlagHasColors)
	{
	    display->ncolors = COLOR_PAIRS;
	    display->color_intensity_max = 0;

	    if(display->ncolors > 0)
	    {
		const int m = COLORS;
		int i, fg_i, bg_i;
		CColor *c;

		/* Allocate the colormap to the number of color pairs */
		display->colormap = (CColor *)malloc(
		    display->ncolors * sizeof(CColor)
		);

		/* Reset the first color, it is never used in curses
		 * for portability reasons
		 */
		c = &display->colormap[0];
		memset(c, 0x00, sizeof(CColor));

		fg_i = 0;
		bg_i = -1;		/* Use default color -1 for the
					 * background for the first
					 * set of colors */

		/* Initialize color pairs 1 to (COLOR_PAIRS - 1) or
		 * whenever we run out of unique fg and bg colors
		 */
		for(i = 1; i < display->ncolors; i++)
		{
		    c = &display->colormap[i];

		    c->fg_i = fg_i;
		    c->bg_i = bg_i;

		    color_content(
			fg_i,
			&c->fg_red, &c->fg_green, &c->fg_blue
		    );

		    if(bg_i > -1)
		    {
			color_content(
			    bg_i,
			    &c->bg_red, &c->bg_green, &c->bg_blue
			);
		    }
		    else
		    {
			/* Clear */
			c->bg_red = -1;
			c->bg_green = -1;
			c->bg_blue = -1;
		    }

		    init_pair(i, c->fg_i, c->bg_i);

		    if(c->fg_red > display->color_intensity_max)
			display->color_intensity_max = c->fg_red;
		    if(c->fg_green > display->color_intensity_max)
			display->color_intensity_max = c->fg_green;
		    if(c->fg_blue > display->color_intensity_max)
			display->color_intensity_max = c->fg_blue;

		    if(c->bg_red > display->color_intensity_max)
			display->color_intensity_max = c->bg_red;
		    if(c->bg_green > display->color_intensity_max)
			display->color_intensity_max = c->bg_green;
		    if(c->bg_blue > display->color_intensity_max)
			display->color_intensity_max = c->bg_blue;   

		    if(fg_i > m)
		    {
			fg_i = 0;
			bg_i++;
			if(bg_i > m)
			{
			    i++;
			    break;
			}
		    }
		    else
		    {
			fg_i++;
		    }
		}

		for(; i < display->ncolors; i++)
		{
		    c = &display->colormap[i];
		    memset(c, 0x00, sizeof(CColor));
		}
	    }
	    else
	    {
		display->colormap = NULL;
	    }
	}
	else
	{
	    display->colormap = NULL;
	    display->ncolors = 0;
	    display->color_intensity_max = 0;
	}

	/* Set root window ID */
	display->root_win = (Window)w;

	/* Get width and height of root window */
	display->root_width = CWRootWidth(display);
	display->root_height = CWRootHeight(display);

#if 0
	/* Set screen */
	s = getenv("TERM");
	if(s != NULL)
	{
	    screen = newterm(
		s,
		stdout,
		stdin
	    );
	}
#endif

	/* Move cursor to root window */
	display->cursor_window = display->root_win;
	display->cursor_x = 0;
	display->cursor_y = 0;


	/* Turn off echo */
	noecho();

	/* Pass key presses directly to us (including interrupts) */
	raw();

	/* Allow line feed */
	nonl();

	/* Enter non-blocking mode */
	/* nodelay(stdscr, TRUE); */
	timeout(0);

	/* Enable keypad and cursor keys */
	keypad(stdscr, TRUE);

	/* Insert and delete key use is ok */
	idcok(stdscr, TRUE);


	return(display);
}

/*
 *	Shutsdown curses library and deallocates display structure.
 */
void CWShutdown(CDisplay *display)
{
	if(display == NULL)
	    return;

	/* Delete the colormap */
	free(display->colormap);
	display->colormap = NULL;
	display->ncolors = 0;

	clrtoeol();	/* Clear bottom line */
	refresh();	/* Flush out everything */
	endwin();	/* Curses clean up */

	/* Delete display structure */
	free(display);
}


/*
 *	Creates a new window.
 *
 *	Returns the window ID or None on error.
 */
Window CWCreateWindow(
	CDisplay *display,
	Window parent,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	return((Window)subwin(
	    (WINDOW *)parent,
	    height, width,
	    y, x
	));
}

/*
 *	Macros to fetch width and height of root window.
 */
unsigned int CWRootWidth(CDisplay *display)
{
#if 1
	return(COLS);
#else
	const char *s = getenv("COLUMNS");
	return((unsigned int)(
	    (s != NULL) ? atoi(s) : DefaultRootWidth
	));
#endif
}

unsigned int CWRootHeight(CDisplay *display)
{
#if 1
	return(LINES);
#else
	const char *s = getenv("LINES");
	return((unsigned int)(
	    (s != NULL) ? atoi(s) : DefaultRootHeight
	));
#endif
}


/*
 *	Returns the color index that most closely matches the specified
 *	criteria.
 *
 *	All values must be in coefficient from 0.0 to 1.0, or -1.0 may
 *	be specified to indicate default/clear.
 *
 *	Can return -1 of no color was matched.
 */
int CWMatchColorCoeff(
	CDisplay *display,   
	const float fg_red, const float fg_green, const float fg_blue,
	const float bg_red, const float bg_green, const float bg_blue
)
{
	const short tolorance = 100;	/* In 0 to 1000 range */
	short	v_max,
		v_fg_red, v_fg_green, v_fg_blue,
		v_bg_red, v_bg_green, v_bg_blue;
	int i;
	CColor *c;

	if(display == NULL)
	    return(-1);

	v_max = display->color_intensity_max;
	if(v_max < 0)
	    v_max = 1000;

	v_fg_red = CLIP((short)(fg_red * v_max), -1, v_max);
	v_fg_green = CLIP((short)(fg_green * v_max), -1, v_max);
	v_fg_blue = CLIP((short)(fg_blue * v_max), -1, v_max);
	v_bg_red = CLIP((short)(bg_red * v_max), -1, v_max);
	v_bg_green = CLIP((short)(bg_green * v_max), -1, v_max);
	v_bg_blue = CLIP((short)(bg_blue * v_max), -1, v_max);

	/* Find the color */
	for(i = 1; i < display->ncolors; i++)
	{
	    c = &(display->colormap[i]);

#define CHECK_IN_RANGE(_ci_,_c_)	(	\
 ((_c_) == -1) || (				\
  ((_ci_) >= ((_c_) - tolorance)) &&		\
  ((_ci_) <= ((_c_) + tolorance))		\
 )						\
)

	    if(!CHECK_IN_RANGE(c->fg_red, v_fg_red))
		continue;
	    if(!CHECK_IN_RANGE(c->fg_green, v_fg_green))
		continue;
	    if(!CHECK_IN_RANGE(c->fg_blue, v_fg_blue))
		continue;

	    if(!CHECK_IN_RANGE(c->bg_red, v_bg_red))
		continue;
	    if(!CHECK_IN_RANGE(c->bg_green, v_bg_green))
		continue;
	    if(!CHECK_IN_RANGE(c->bg_blue, v_bg_blue))
		continue;

	    return(i);
	}

	return(-1);
}


/*
 *	Sets the color for the specified window.
 */
void CWSetColor(
	CDisplay *display, Window w,
	const int color_num, const unsigned int flags
)
{
	CColor *c;

	if((display == NULL) || (w == None))
	    return;  

	/* No colormap or colors not supported? */
	if(display->colormap == NULL)
	    return;

	c = ((color_num >= 1) && (color_num < display->ncolors)) ?
	    &(display->colormap[color_num]) : NULL;
	if(c == NULL)
	    return;

	wcolor_set((WINDOW *)w, (short)color_num, NULL);

	if(flags & CColorFlagBold)
	    wattron((WINDOW *)w, A_BOLD);
	else
	    wattroff((WINDOW *)w, A_BOLD);

	if(flags & CColorFlagDim)
	    wattron((WINDOW *)w, A_DIM);
	else
	    wattroff((WINDOW *)w, A_DIM);

	if(flags & CColorFlagReverse)
	    wattron((WINDOW *)w, A_REVERSE);
	else
	    wattroff((WINDOW *)w, A_REVERSE);

	if(flags & CColorFlagUnderline)
	    wattron((WINDOW *)w, A_UNDERLINE);
	else
	    wattroff((WINDOW *)w, A_UNDERLINE);

	if(flags & CColorFlagBlink)
	    wattron((WINDOW *)w, A_BLINK);
	else
	    wattroff((WINDOW *)w, A_BLINK);

}

/*
 *	Moves the cursor position to the Window w and to a specified
 *	position.
 */
void CWMoveCursor(
	CDisplay *display, Window w,
	const int x, const int y
)
{
	if((display == NULL) || (w == None))
	    return;

	/* Move cursor to window and specified position */
	wmove((WINDOW *)w, y, x);

	/* Update current cursor location */
	display->cursor_window = w;
	display->cursor_x = x;
	display->cursor_y = y;
}

/*
 *	Draws a character at the specified position on the Window
 *	w.
 */
void CWDrawCharacter(
	CDisplay *display, Window w,
	const int x, const int y,
	char c
)
{
	if((display == NULL) || (w == None))
	    return;

	mvwaddch((WINDOW *)w, y, x, c);
}

/*
 *	Draws a string at the current cursor position on the Window
 *	w.
 *
 *	If there are any '%' characters in the string, then each one
 *	must be doubled. Example "Progress is 52%% complete.".
 */
void CWDrawString(
	CDisplay *display, Window w,
	const char *s
)
{
	int len;

	if((display == NULL) || (w == None) || (s == NULL))
	    return;

	/* Get length of string */
	len = strlen(s);

	/* Draw string */
	wprintw((WINDOW *)w, (char *)s);

	/* Update current cursor location */
	display->cursor_window = w;
	display->cursor_x += len;
}

/*
 *	Draws a string at the specified cursor position on the Window
 *      w.
 *
 *      If there are any '%' characters in the string, then each one
 *      must be doubled. Example "Progress is 52%% complete.".
 */
void CWDrawStringPosition(
	CDisplay *display, Window w,
	const int x, const int y,
	const char *s
)
{
	int i, xc;

	if((display == NULL) || (w == None) || (s == NULL))
	    return;

	for(xc = x, i = 0; s[i] != '\0'; xc++, i++)
	    mvwaddch(
		(WINDOW *)w,
		y, xc,
		s[i]
	    );

	/* Update current cursor location */
	display->cursor_window = w;
	display->cursor_x = x + i;
	display->cursor_y = y;
}

/*
 *	Same as CWDrawStringPosition() except that it will not draw
 *	more than the number of characters specified by limit.
 */
void CWDrawStringPositionLimited(
	CDisplay *display, Window w,
	const int x, const int y,
	const char *s,
	const int limit
)
{
	int i, xc;

	if((display == NULL) || (w == None) || (s == NULL))
	    return;

	for(xc = x, i = 0; s[i] != '\0'; xc++, i++)
	{
	    if(i >= limit)
		break;

	    mvwaddch(
		(WINDOW *)w,
		y, xc,
		s[i]
	    );
	}

	/* Update current cursor location */
	display->cursor_window = w;
	display->cursor_x = x + i;
	display->cursor_y = y;
}


/*
 *	Puts ' ' characters to the specified area. The specified
 *	width and height must be atleast one or greater.
 */
void CWClearArea(
	CDisplay *display, Window w,
	const int x, const int y,
	const unsigned int width, const unsigned int height
)
{
	int xc, yc, wc, hc;
	char c = ' ';

	if((display == NULL) || (w == None) ||
	   (width == 0) || (height == 0)
	)
	    return;

	wc = x + (int)width;
	hc = y + (int)height;

	for(yc = y; yc < hc; yc++)
	{
	    for(xc = x; xc < wc; xc++)
	        mvwaddch(
		    (WINDOW *)w,
		    yc, xc,
		    c
		);
	}

	/* Update current cursor location */
	display->cursor_window = w;
	display->cursor_x = x + (int)width;
	display->cursor_y = y + (int)height;
}


/*
 *	Fetches the next event, returns positive if it got an event,
 *	or 0 if there were no events or an error occured.
 */
int CGetNextEvent(
	CDisplay *display,
	CEvent *event
)
{
	int c, got_event = 0;

	if((display == NULL) || (event == NULL))
	    return(got_event);

	/* Check for CKeyPress */
	c = getch();
	if(c != ERR)
	{
	    event->type = CKeyPress;
	    event->ckey.key = c;
	    got_event++;
	}

	refresh();		/* Flush out everything */

	return(got_event);
}


/*
 *	Syncronizes the curses library.
 */
void CWSync(CDisplay *display, const unsigned int options)
{
	if(display == NULL)
	    return;

	refresh();
}
