#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/unvfile.h"

#include "net.h"
#include "xsw.h"


int DBIsObjectGarbage(int obj_num);

static xsw_object_struct **DBReallocObjectList(
	xsw_object_struct **list, int total
);
static xsw_object_struct *DBObjectNew(void);
static void DBObjectDelete(xsw_object_struct *obj);

int DBCreateObject(
	int imgref_num,
	int type,
	int owner,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank
);
int DBCreateExplicitObject(
	int obj_num,
	int imgref_num,
	int type,
	int owner,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank
);

void DBDeleteAllObjects(void);
void DBRecycleObject(int obj_num);

void DBReclaim(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *      Checks if object number is valid, allocated and garbage.
 *
 *      Returns false if object is valid, allocated and non-garbage.
 */
int DBIsObjectGarbage(int obj_num)
{
	return(
	    UNVIsObjectGarbage(
		xsw_object,
		total_objects,
		obj_num
	    )
	);
}

/*
 *	Reallocates the objects list.
 */
static xsw_object_struct **DBReallocObjectList(
	xsw_object_struct **list, int total
)
{
	if(total > 0)
	{
	    list = (xsw_object_struct **)realloc(
		list,
		total * sizeof(xsw_object_struct *)
	    );
	}
	return(list);
}

/*
 *	Creates a new object.
 *
 *	It's initial values will be set to that of of the global
 *	garbage object.
 */
xsw_object_struct *DBObjectNew()
{
	xsw_object_struct *obj = (xsw_object_struct *)calloc(
	    1, sizeof(xsw_object_struct)
	);
	if(obj == NULL)
	    return(NULL);
 
	/* Reset values on new object */
	UNVResetObject(obj);

	return(obj);
}

/*
 *	Deletes the object.
 */
static void DBObjectDelete(xsw_object_struct *obj)
{
	xsw_net_parms_struct *net = &net_parms;

	if(obj == NULL)
	    return;

	/* If this is the player object being deleted then all
	 * references to the player object must be reset
	 */
	if(obj == net->player_obj_ptr)
	    DBSetPlayerObject(-1);

	/* Delete object from in range list */
	DBInRangeDelete(obj);

	/* Delete object and all its resources */
	UNVDeleteObject(obj);
}


/*      
 *	Creates a new object in the database.
 *
 *	Returns the index of the new object or -1 on error.
 */
int DBCreateObject(
	int imgref_num,
	int type,
	int owner,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank
)
{
	int i, n, prev_total;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* Invalid object type? */
	if(type <= SW_OBJ_TYPE_GARBAGE)
	{
	    fprintf(
		stderr,
"DBCreateObject: Error: Useless request to create garbage.\n"
	    );
	    return(-1);
	}

	/* Is the specified imgref number valid? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return(-1);

	/* Load the imgref as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);

	/* Look for an available garbage object in the database */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type == SW_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < total_objects)
	{
	    /* Got available object */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
#ifndef XSW_OBJECT_ALLOCATE_AHEAD
# define XSW_OBJECT_ALLOCATE_AHEAD	10
#endif
	    /* Can we allocate more? */
	    if((total_objects + XSW_OBJECT_ALLOCATE_AHEAD) > SW_OBJECTS_MAX)
		return(-1);

	    prev_total = total_objects;
	    total_objects += XSW_OBJECT_ALLOCATE_AHEAD;

	    /* Allocate pointer array */
	    xsw_object = DBReallocObjectList(
		xsw_object, total_objects
	    );
	    if(xsw_object == NULL)
	    {
		total_objects = 0;
		return(-3);
	    }

	    /* Allocate each object */
	    for(i = prev_total;
		i < total_objects;
		i++
	    )
	    {
		xsw_object[i] = DBObjectNew();
		if(xsw_object[i] == NULL)
		{
		    total_objects = i;
		    return(-1);
		}
	    }

	    /* New object will be the first object allocated */
	    n = prev_total;
	}

	/* Get available object and reset its values */
	obj = xsw_object[n];

	UNVResetObject(obj);
	obj->type = type;
	obj->owner = owner;
	obj->imgref = imgref_num;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	MuSetUnitVector2D(&obj->attitude_vector_compoent, obj->heading);

	obj->last_updated = cur_millitime;
	obj->birth_time = cur_millitime;

	/* Add this object to the in range objects list */
	DBInRangeAdd(
	    net->player_obj_num,
	    n,
	    0
	);

	return(n);
}

/*
 *	Creates a new object in the database explicitly by the specified
 *	index obj_num.
 *
 *	If the object is non-garbage then it will be recycled first.
 *
 *	Returns non-zero on error.
 */
int DBCreateExplicitObject(
	int obj_num,
	int imgref_num,
	int type,
	int owner,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank
)
{
	char *s;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	if((type <= SW_OBJ_TYPE_GARBAGE) || (obj_num < 0))
	    return(-2);

	/* Is the specified imgref number invalid? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Error: Invalid imgref number (imgref_num=%i).\n",
		imgref_num
	    );
	    return(-2);
	}

	/* If the object exists and is non-garbage then recycle it
	 * first
	 */
	if(!DBIsObjectGarbage(obj_num))
	    DBRecycleObject(obj_num);

	/* Load the imgref's image as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);

	/* Allocate more object pointers as needed */
	if(obj_num >= total_objects)
	{
	    const int prev_total = MAX(total_objects, 0);
	    int i;

	    total_objects = obj_num + 1;
	    xsw_object = DBReallocObjectList(
		xsw_object,
		total_objects
	    );
	    if(xsw_object == NULL)
	    {
		total_objects = 0;
		return(-3);
	    }
	    for(i = prev_total; i < total_objects; i++)
		xsw_object[i] = DBObjectNew();
	}

	obj = xsw_object[obj_num];
	if(obj == NULL)
	    return(-3);

	s = (char *)malloc(40 + 80);
	sprintf(s, "Object %i", obj_num);
	strncpy(obj->name, s, sizeof(obj->name));
	obj->name[sizeof(obj->name) - 1] = '\0';
	free(s);

	obj->type = type;
	obj->owner = owner;
	obj->imgref = imgref_num;
	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	MuSetUnitVector2D(&obj->attitude_vector_compoent, obj->heading);

	obj->last_updated = cur_millitime;
	obj->birth_time = cur_millitime;

	/* Add this object to the in range objects list */
	DBInRangeAdd(
	    net->player_obj_num,
	    obj_num,
	    0
	);

	return(0);
}


/*
 *	Deletes all objects in the database, deletes the inrange objects
 *	list, viewscreen labels, scanner contacts, and resets all
 *	references to the player object.
 */
void DBDeleteAllObjects(void)
{
	int i;
	xsw_object_struct *obj;

	/* Go through XSW objects list */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    DBObjectDelete(obj);
	}
	free(xsw_object);
	xsw_object = NULL;
	total_objects = 0;

	/* Delete the in range objects list */
	DBInRangeDeleteAll();

	/* Reset all player object references */
	DBSetPlayerObject(-1);

	/* Delete all viewscreen labels */
	VSLabelDeleteAll();

	/* Delete all scanner contacts */
	ScDeleteAll();
}


/*
 *	Recycles the object.
 *
 *	All its values will be reset and any allocated substructures
 *	will be deleted.
 *
 *	Associated viewscreen label and its entry in the inrange
 *	objects list will be deleted as well. If this object is
 *	the player object, then all referances to the player object
 *	will be reset.
 */
void DBRecycleObject(int obj_num)
{
	int i;
	xsw_object_struct *obj, *tar_obj;
	xsw_net_parms_struct *net = &net_parms;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Begin removing all references to this object */

	/* Iterate through the objects list and remove any references
	 * to this object by other objects
	 */
	for(i = 0; i < total_objects; i++)
	{
	    if(i == obj_num)
		continue;

	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    if(tar_obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Tractored objects */
	    if(tar_obj->total_tractored_objects > 0)
	    {
		const int m = tar_obj->total_tractored_objects;
		int n;
		for(n = 0; n < m; n++)
		{
		    if(tar_obj->tractored_object[n] == obj_num)
			tar_obj->tractored_object[n] = -1;
		}
	    }
	}

	/* If this is the player object then reset all references to
	 * it
	 */
	if(obj_num == net->player_obj_num)
	    DBSetPlayerObject(-1);

	/* If this is the object that the player object is locked on
	 * then have the player object unlock
	 */
	if(net->player_obj_ptr != NULL)
	{
	    xsw_object_struct *player_obj = net->player_obj_ptr;
	    if(player_obj->locked_on > -1)
	    {
		if(player_obj->locked_on == obj_num)
		{
		    player_obj->locked_on = -1;

		    BridgeWinDrawPanel(
			-1,
			BW_PANEL_S1
		    );
		    BridgeWinDrawPanel(
			-1,
			BW_PANEL_S2
		    );
		    BridgeWinDrawPanel(
			-1,
			BW_PANEL_S3
		    );
		}
	    }
	}

	/* Delete viewscreen labels refering to this object */
	VSLabelDeleteByObjectPtr(obj);

	/* Remove object from the inrange objects list */
	DBInRangeDelete(obj);

	/* Reset values and delete all sub structures */
	UNVResetObject(obj);

	/* Mark this object as recycled by setting its type to
	 * SW_OBJ_TYPE_GARBAGE
	 */
	obj->type = SW_OBJ_TYPE_GARBAGE;
}


/*
 *      Frees outdated objects and reallocates the objects list.
 */
void DBReclaim(void)
{
	int i, h;
	time_t t;
	xsw_object_struct *obj, *player_obj;
	xsw_net_parms_struct *net = &net_parms;

	/* Sanitize total */
	if(total_objects < 0)
	    total_objects = 0;

	/* No objects allocated? */
	if((total_objects == 0) || (xsw_object == NULL))
	    return;

	/* Get highest index numbered object that is valid */
	for(i = 0, h = -1; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    h = i;
	}


	/* Delete all allocated objects from total_objects - 1 to h + 1.
	 *
	 * The global player object referance will also be reset to -1 if
	 * it is deleted in the call to DBObjectDelete(), but this
	 * probably won't happen since it would only be deleted if it was
	 * type garbage or invalid.
	 */
	for(i = total_objects - 1; i > h; i--)
	    DBObjectDelete(xsw_object[i]);

	/* Adjust global total_objects */
	total_objects = h + 1;
	if(total_objects > 0)
	{
	    /* Reallocate pointers */
	    xsw_object = DBReallocObjectList(
		xsw_object,
		total_objects
	    );
	    if(xsw_object == NULL)
	    {
		total_objects = 0;
	    }
 	}
	else
	{
	    /* Delete all pointers */
	    free(xsw_object);
	    xsw_object = NULL;
	    total_objects = 0;
	}
  

	/* Get pointer to player object */
	player_obj = net->player_obj_ptr;

	/* Go through the objects list, begin sanitizing them
	 *
	 * Some objects may need to be recycled
	 */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Skip the player object */
	    if(obj == player_obj)
		continue;

	    /* If object has not been updated recently, then recycle
	     * it
	     */
	    t = obj->last_updated + option.net_object_outdated_time;
	    if(t < cur_millitime)
		DBRecycleObject(i);

	    /* If object is not in the same sector with the player
	     * then recycle it
	     */
	    if(!Mu3DInSameSectorPtr(obj, player_obj))
		DBRecycleObject(i);
	}
}
