#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/imgrefs.h"
#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/unvmain.h"
#include "../include/unvutil.h"
#include "../include/unvmath.h"

#include "../include/swsoundcodes.h"
#include "../include/swnetcodes.h"

#include "ocsinfofio.h"
#include "particle.h"
#include "sound.h"
#include "net.h"
#include "swterm.h"
#include "ocsnames.h"


static void SWTSignalCB(int s);
void SWTResetTimmers(const time_t cur_ms, const time_t cur_sec);
void SWTReclaimGlobalMemory(bool_t print_response);

void SWTOpenOCSNs(const char *path);
void SWTOpenImgRefs(const char *path);
void SWTOpenSndRefs(const char *path);

void SWTUpdateBackgroundMusic(void);

int SWTConnect(const char *arg);
void SWTDisconnect(void);
void SWTRefresh(void);

void SWTDamageNotify(
	const int obj_num,
	const float prev_hp, const float hp,
	const float prev_power, const float power,
	const int shield_state
);
void SWTHitNotify(
	const int src_obj_num,
	const int tar_obj_num,
	const int owner_obj_num,
	const float dmg_hull,
	const float dmg_shield,
	const float bearing
);
void SWTDestroyNotify(
	const int src_obj_num,
	const int tar_obj_num,
	const int owner_obj,
	const int reason
);

int SWTManageGUI(void *data);
int SWTManageSound(void *data);
int SWTManageNet(void *data);

int SWTInit(int argc, char **argv);
void SWTMainIteration(void *data);
void SWTMain(void *data);
void SWTMainPop(void *data);
void SWTMainPopAll(void *data);
void SWTShutdown(void);


int		runlevel;

time_t		cur_millitime;
time_t		cur_systime;
time_t		lapsed_millitime;
float		time_compensation;

xsw_debug_struct	debug;
xsw_option_struct	option;
osw_gui_struct	osw_gui[1];
CDisplay	**curses_display;
int		total_curses_displays;
xsw_dname_struct	dname;
xsw_fname_struct	fname;
xsw_next_struct	next;

xsw_player_values_struct	player_values;
xsw_gctl_struct		*xsw_gctl;

sw_legend_struct	sw_legend;

xsw_scanner_contact_struct	**scanner_contact;
int				total_scanner_contacts;

xsw_ocsn_struct	**xsw_ocsn;
int		total_ocsns;

xsw_net_stat_struct	net_stat;
xsw_auto_interval_tune_struct	auto_interval_tune;

xsw_font_struct	xsw_font;
swt_color_struct	swt_color;

swt_bridge_struct	*swt_bridge;
swt_eco_win_struct	*xsw_ecow;

dialog_struct	err_dw;
dialog_struct	info_dw;

imgref_struct	**xsw_imgref;
int		total_imgrefs;

xsw_object_struct	**xsw_object;
int			total_objects;

xsw_object_struct	**inrange_xsw_object;
int			total_inrange_objects;

xsw_particle_struct	**xsw_particle;
int			total_particles;

xsw_net_parms_struct	net_parms;

xsw_sound_struct	sound;
sndref_struct	**xsw_sndref;
int		total_sndrefs;


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Signal handler.
 */
static void SWTSignalCB(int s)
{
	switch(s)
	{
	  case SIGHUP:
/* TODO reload configuration */
	    SWTRefresh();
	    break;

	  case SIGINT:
	    SWTMainPopAll(NULL);
	    break;

	  case SIGQUIT:
	    SWTMainPopAll(NULL);
	    break;

	  case SIGABRT:
	    SWTMainPopAll(NULL);
	    break;

	  case SIGKILL:
	    SWTMainPopAll(NULL);
	    break;

	  case SIGPIPE:
	    break;

	  case SIGSEGV:
	    XSWPrintDebugStderr(1);
	    exit(1);
	    break;

	  case SIGTERM:
	    SWTMainPopAll(NULL);
	    break;

	  case SIGSTOP:
	    break;

	  case SIGCONT:
	    break;

	  default:
	    if(s > 0)
	    {
		char *msg = (char *)malloc(40 + 80);
		sprintf(
		    msg,
"Received unknown signal `%i'.\n",
		    s
		);
		XSWPrintWarning(msg);
		free(msg);
	    }
	    break;
	}
}

/*
 *	Resets global timmers.
 */
void SWTResetTimmers(const time_t cur_ms, const time_t cur_sec)
{
	int i, n;
	xsw_object_struct *obj;
	xsw_weapon_struct *wep;

	next.update_check = cur_ms;
	next.viewscreen = cur_ms;
	next.bridge_std_redraw = cur_ms;
	next.memory_clean = cur_ms;
	next.lplayer_pos_send = cur_ms;
	next.spawn_particles = cur_ms;
	next.wormhole_enter = cur_ms;

	auto_interval_tune.next = 0l;

	/* Objects */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    obj->last_updated = cur_ms;
	    if(obj->birth_time > 0l)
		obj->birth_time = cur_ms;
	    if(obj->last_cloak > 0l)
		obj->last_cloak = cur_ms;
	    obj->animation.last_interval = cur_ms;

	    /* Auto destruct needs to be unset since there is no way
	     * to calculate it after the timmers being reset
	     */
	    obj->auto_destruct_next = 0l;

	    /* Last animation frame */
	    obj->animation.last_interval = 0l;

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		wep->last_used = 0l;
	    }
	}
}


/*
 *	Reclaims global memory.
 */
void SWTReclaimGlobalMemory(bool_t print_response)
{
	xsw_net_parms_struct *net = &net_parms;

	if(print_response)
	    XSWPrint("Reclaiming memory...");

	DBReclaim();
/* No imgrefs to reclaim */
/* No vslabels reclaim */
	DBInRangeUpdate(net->player_obj_num);

	if(print_response)
	    XSWPrint("Memory reclaimed.");
}

/*
 *	Opens the OCS Names file.
 */
void SWTOpenOCSNs(const char *path)
{
	int i, total;
	char *full_path;
	ocs_info_struct **list;
	xsw_net_parms_struct *net = &net_parms;

	if(STRISEMPTY(path))
	    return;

	full_path = ISPATHABSOLUTE(path) ?
	    STRDUP(path) :
	    STRDUP(PrefixPaths(dname.etc, path));

	/* Delete all the existing OCS Names */
	OCSNClear(&xsw_ocsn, &total_ocsns);

	/* Open the OCS Names from file */
	list = NULL;
	total = 0;
	if(!OCSInfoFileOpen(
	    &list, &total, full_path,
	    NULL, NULL,
	    XSWPrintWarning
	))
	{
	    int ocsn_num;
	    ocs_info_struct *ocs_info;
	    xsw_ocsn_struct *ocsn;

	    /* Parse the loaded OCS Info list and convert each OCS
	     * Info to an OCS Name
	     */
	    for(i = 0; i < total; i++)
	    {
		ocs_info = list[i];
		if(ocs_info == NULL)
		    continue;

		/* Create a new OCS Name for this OCS Info */
		ocsn_num = MAX(total_ocsns, 0);
		total_ocsns = ocsn_num + 1;
		xsw_ocsn = (xsw_ocsn_struct **)realloc(
		    xsw_ocsn,
		    total_ocsns * sizeof(xsw_ocsn_struct *)   
		);
		xsw_ocsn[ocsn_num] = ocsn = XSW_OCSN(calloc(
		    1, sizeof(xsw_ocsn_struct)
		));
		if(ocsn == NULL)
		    break;

		ocsn->code = ocs_info->code;
		ocsn->name = STRDUP(ocs_info->name);

	    }

	    strncpy(fname.ocsn, full_path, sizeof(fname.ocsn));
	    fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';

	    /* Updated resources that would be affected by this change */

	    /* Remove weapons from player object so they get recreated */
	    UNVAllocObjectWeapons(
		net->player_obj_ptr,
		0
	    );

	    /* Recreate the selected weapon viewscreen label */
	    if(swt_bridge != NULL)
		VSDrawUpdateWeaponLabel(
		    &swt_bridge->vs_weapon_image,
		    swt_bridge->vs_weapon_buf
		);
	}

	/* Delete all the loaded OCS Infos */
	for(i = 0; i < total; i++)
	    OCSInfoDelete(list[i]);
	free(list);

	free(full_path);
}

/*
 *	Opens the ImgRefs file.
 */
void SWTOpenImgRefs(const char *path)
{

}

/*
 *	Opens the SndRefs file.
 */
void SWTOpenSndRefs(const char *path)
{

}


/*
 *	Updates the background music.
 *
 *	This should be called whenever the background music might
 *	need changing.
 */
void SWTUpdateBackgroundMusic(void)
{
	int code = SW_SNDREF_BG_MUSIC_STANDARD;
	xsw_object_struct *obj_ptr, *tar_obj_ptr;
	xsw_net_parms_struct *net = &net_parms;
	swt_bridge_struct *bw = swt_bridge;

	if(bw == NULL)
	    return;

	/* Check if main menu on bridge window is mapped */
	if(bw->screen_type == BRIDGE_SCREEN_TYPE_MAIN_MENU)
	{
	    /* Main menu on bridge window is mapped */

	    code = SW_SNDREF_BG_MUSIC_MENU_MAIN;
	}
	else
	{
	    /* View screen is mapped */

	    code = SW_SNDREF_BG_MUSIC_STANDARD;

	    if(net->player_obj_ptr == NULL)
	    {
		/* Player object not valid, could be still logging in */

	    }
	    else
	    {
		/* Player object is valid */

		/* Get pointer to player object */
		obj_ptr = net->player_obj_ptr;

		/* Check if in nebula */
		if(obj_ptr->loc_type == SW_LOC_TYPE_NEBULA)
		    code = SW_SNDREF_BG_MUSIC_NEBULA;
		else if(obj_ptr->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
		    code = SW_SNDREF_BG_MUSIC_ASTEROID_FIELD;
#if 0
		else if(obj_ptr->loc_type == SW_LOC_TYPE_PULSAR)
		    code = SW_SNDREF_BG_MUSIC_PULSAR;
#endif

		/* Check if locked onto something */
		if(DBIsObjectGarbage(obj_ptr->locked_on))
		{
		    /* Player object is not locked on anything */

		}
		else
		{
		    /* Player object is locked on something */
		    tar_obj_ptr = xsw_object[obj_ptr->locked_on];

		    /* Check if locked on target is controlled,
		     * weapon, or player.
		     */
		    if((tar_obj_ptr->type == SW_OBJ_TYPE_CONTROLLED) ||
		       (tar_obj_ptr->type == SW_OBJ_TYPE_PLAYER) ||
		       (tar_obj_ptr->type == SW_OBJ_TYPE_WEAPON) ||
		       (tar_obj_ptr->type == SW_OBJ_TYPE_STREAMWEAPON) ||
		       (tar_obj_ptr->type == SW_OBJ_TYPE_SPHEREWEAPON)
		    )
			code = SW_SNDREF_BG_MUSIC_ENGAGING;


		}
	    }
	}

	/* Change background music (as needed) */
	if(code != sound.bg_music_sndref_num)
	    SoundChangeBackgroundMusic(&sound, code, 0, 0);
}


/*
 *	Connects to the universe specified by url.
 */
int SWTConnect(const char *url)
{
	int status;
	char *s, *durl;
	xsw_net_parms_struct *net = &net_parms;

	if(STRISEMPTY(url))
	    return(-2);

	durl = STRDUP(url);
	if(durl == NULL)
	    return(-3);

	strstrip(durl);

	/* Disconnect if already connected */
	SWTDisconnect();

	/* Get the address */
	s = StringParseAddress(durl);
	if(STRISEMPTY(s))
	{
	    char *s = (char *)malloc(strlen(durl) + 80);
	    sprintf(
		s,
"Unable to parse url: %s.",
		durl
	    );
	    XSWPrintWarning(s);
	    free(s);
	    free(durl);
	    return(-2);
	}
	free(net->address);
	net->address = STRDUP(s);

	/* Get the port */
	net->port = StringParsePort(durl);
	if(net->port < 0)
	    net->port = XSW_DEF_SERVER_PORT;

	/* Get the login name */
	s = StringParseName(durl);
	if(STRISEMPTY(s))
	{
	    free(net->login_name);
	    net->login_name = STRDUP(XSW_DEF_GUEST_LOGIN_NAME);
	}
	else
	{
	    free(net->login_name);
	    net->login_name = STRDUP(s);
	}

	/* Get the login password */
	s = StringParsePassword(durl);
	if(STRISEMPTY(s))
	{
	    free(net->login_password);
	    net->login_password = STRDUP(XSW_DEF_GUEST_LOGIN_PASSWORD);
	}
	else
	{
	    free(net->login_password);
	    net->login_password = STRDUP(s);
	}

	/* Connect */
	status = NetConnect(net->address, net->port, 45l);
	if(status < 0)
	{
	    /* Connect failed, NetConnect() will print error messages */
	}
	else
	{
	    char *s = (char *)malloc(STRLEN(net->address) + 80);
	    sprintf(
		s,
		"Connecting to %s...",
		net->address
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Redraw */
	BridgeDraw(swt_bridge, 0);

	free(durl);

	return(status);
}

/*
 *	Disconnects from the universe.
 */
void SWTDisconnect(void)
{
	bool_t deallocate_resources = False;
	xsw_net_parms_struct *net = &net_parms;

	/* Check if currently connected */
	if(net->socket > -1)
	{
	    if(XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    {
		NetWaitClose(45l);
	    }
	    else
	    {
		NetSendLogout();
		NetShutdownConnection(XSW_CON_SHUTDOWN_RDWR, 1);
		NetWaitClose(45l);
	    }
	    NetDisconnect(45l);
	    deallocate_resources = True;
	}
	else
	{
	    /* Not connected, reset all network resources and deallocate
	     * other resources
	     */
	    deallocate_resources = True;
	}

	/* Deallocate all network related resources? */
	if(deallocate_resources)
	{
	    xsw_gctl_struct *gctl = xsw_gctl;

	    /* Delete all objects. This will also reset the player
	     * object referance, delete all scanner contacts, and
	     * reset inrange objects list
	     */
	    DBDeleteAllObjects();
	    ParticleClear(&xsw_particle, &total_particles);

	    /* Reset game controller positions */
	    if(gctl != NULL)
	    {
		gctl->turn = 0.0f;
		gctl->throttle = 0.0f;
		gctl->thrust_dir = 0.0f * PI;
		gctl->fire_weapon = False;
		gctl->omni_dir_thrust = False;
		gctl->external_dampers = False;
	    }
	}

	/* Completely redraw bridge window */
	BridgeDraw(swt_bridge, 0);
}

/*
 *      Does a network refresh sequence.
 */
void SWTRefresh(void)
{
	xsw_net_parms_struct *net = &net_parms;

#if 0
	/* Delete object labels */
	VSLabelDeleteAll();
#endif

	/* Delete all scanner contacts */
	ScDeleteAll();

	/* Update the inrange objects list */
	DBInRangeUpdate(net->player_obj_num);


	/* Currently connected? */
	if((net->socket > -1) && XSW_CON_STATE_IS_CONNECTED(net))
	{
	    /* Send refresh to server */
	    NetSendRefresh();
	}

	/* Redraw */
	BridgeDraw(swt_bridge, 0);
}

/*
 *	Damage notify.
 *
 *	This function should be called whenever an object has received
 *	damage.
 */
void SWTDamageNotify(
	const int obj_num,
	const float prev_hp, const float hp,
	const float prev_power, const float power,
	const int shield_state
) 
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_object_struct *obj;
	xsw_sound_struct *snd = &sound;
	xsw_option_struct *opt = &option;

	obj = DBIsObjectGarbage(obj_num) ?
	    NULL : xsw_object[obj_num];
	if(obj == NULL)
	    return;

	if(obj_num == net->player_obj_num)
	{
	    bool_t      hp_warning = False,
			hp_danger = False,
			shield_warning = False,
			shield_danger = False;
	    if(obj->hp_max > 0.0f)   
	    {
		const float m = obj->hp_max;
		if(((prev_hp / m) > SWT_HP_WARNING_COEFF) &&
		   ((hp / m) <= SWT_HP_WARNING_COEFF)
		)
		    hp_warning = True;
		if(((prev_hp / m) > SWT_HP_DANGER_COEFF) &&
		   ((hp / m) <= SWT_HP_DANGER_COEFF)
		)
		    hp_danger = True;
	    }
	    if((obj->power_max > 0.0f) &&   
	       (shield_state == SW_SHIELD_STATE_UP)
	    )
	    {
		const float m = obj->power_max;
		if(((prev_power / m) > SWT_POWER_WARNING_COEFF) &&
		   ((power / m) <= SWT_POWER_WARNING_COEFF)       
		)
		    shield_warning = True;
		if(((prev_power / m) > SWT_POWER_DANGER_COEFF) &&
		   ((power / m) <= SWT_POWER_DANGER_COEFF)
		)
		    shield_danger = True;
	    }

	    if(shield_warning)
	    {
		XSWPrintWarning(
"Shields compromised."
		);
		if(opt->sounds >= XSW_SOUNDS_EVENTS)
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
	    }
	    if(shield_danger)
	    {
		XSWPrintWarning(
"Shields failing."
		);
		if(opt->sounds >= XSW_SOUNDS_EVENTS)
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
	    }
	    if(hp_warning)
	    {
		XSWPrintWarning(
"Hull structure compromised."
		);
		if(opt->sounds >= XSW_SOUNDS_EVENTS) 
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
	    }
	    if(hp_danger)
	    {
		XSWPrintError(
"Hull structure failing."
		);
		if(opt->sounds >= XSW_SOUNDS_EVENTS)
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
	    }
	}
}

/*
 *	Hit notify.
 */
void SWTHitNotify(
	const int src_obj_num,
	const int tar_obj_num,
	const int owner_obj_num,
	const float dmg_structure,
	const float dmg_shield,
	const float bearing
)
{
	float vol_left, vol_right;
	int sndref_num = SW_SNDREF_DEFAULT;
	xsw_object_struct       *src_obj, *tar_obj,
				*owner_obj, *camera_obj;
	xsw_net_parms_struct *net = &net_parms;

	/* Get pointers to source and target objects */
	if(DBIsObjectGarbage(src_obj_num))
	    src_obj = NULL;
	else
	    src_obj = xsw_object[src_obj_num];

	if(DBIsObjectGarbage(tar_obj_num))
	    tar_obj = NULL;
	else
	    tar_obj = xsw_object[tar_obj_num];

	if(DBIsObjectGarbage(owner_obj_num))
	    owner_obj = NULL;
	else                 
	    owner_obj = xsw_object[owner_obj_num];

	/* Get the camera object as the player object */
	camera_obj = net->player_obj_ptr;
	if(camera_obj == NULL)                
	    return;

	/* Get distance from camera object to target object */
	if(Mu3DInSameSectorPtr(camera_obj, tar_obj))
	{
	    vol_left = Mu3DDistance(
		tar_obj->x - camera_obj->x,
		tar_obj->y - camera_obj->y,
		tar_obj->z - camera_obj->z
	    );
	    /* Max distance is 3.0 real units */   
	    vol_left = (3.0f - vol_left) / 3.0f;
	    if(vol_left < 0.0f)
		vol_left = 0.0f;

	    vol_right = vol_left;
	}
	else
	{
	    vol_left = 0.0f;
	    vol_right = 0.0f;
	}   


	/* Is the target object the player? */
	if(tar_obj_num == net->player_obj_num)
	{
	    const char *vect_name = DBGetObjectVectorName(bearing);
	    char *s;

	    /* Notify that we were hit */
	    if(owner_obj != NULL)
	    {
		s = (char *)malloc(
		    STRLEN(vect_name) + STRLEN(owner_obj->name) +
		    40 + 40 + (2 * 80)
		);
		sprintf(
		    s,
"Recieved hit to %s from %s: %.2f damage to structure,\
 %.2f damage to shields.",
		    vect_name,
		    owner_obj->name,
		    dmg_structure, dmg_shield
		);
		sndref_num = SW_SNDREF_HIT_STRUCTURE;
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(vect_name) + 40 + 40 + (2 * 80)
		);
		sprintf(
		    s,
"Recieved hit to %s from unknown: %.2f damage to structure,\
 %.2f damage to shields.",
		    vect_name, 
		    dmg_structure, dmg_shield
		);
		sndref_num = SW_SNDREF_HIT_SHIELDS;
	    }
	    XSWPrint(s);
	    free(s);

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    &sound, sndref_num,
		    vol_left, vol_right,
		    0,
		    0
		);
	}
	/* Check if weapon object owner is the player object */
	else if(owner_obj_num == net->player_obj_num)
	{
	    const char *vect_name = DBGetObjectVectorName(bearing);
	    char *s;

	    /* Notify that we hit an object */
	    if(tar_obj == NULL)
	    {
		if(dmg_shield > 0.0f)
		{
		    s = (char *)malloc(
			STRLEN(vect_name) + 40 + 40 + (2 * 80)
		    );
		    sprintf(
			s,
"Unknown object hit %s: %.2f damage to structure,\
 %.2f damage to shields.",
			vect_name, dmg_structure, dmg_shield
		    );
		    sndref_num = SW_SNDREF_OHIT_SHIELDS;
		} 
		else
		{
		    s = (char *)malloc(
			STRLEN(vect_name) + 40 + (2 * 80)
		    );
		    sprintf(
			s,
"Unknown object hit %s: %.2f damage to structure.",
			vect_name, dmg_structure
		    );
		    sndref_num = SW_SNDREF_OHIT_STRUCTURE;
		}
	    }
	    else
	    {
		if(dmg_shield > 0.0f)
		{
		    s = (char *)malloc(
			STRLEN(tar_obj->name) + STRLEN(vect_name) +
			40 + 40 + (2 * 80)
		    );
		    sprintf(
			s,
"%s hit %s: %.2f damage to structure, %.2f damage to shields.",
			tar_obj->name, vect_name,
			dmg_structure, dmg_shield
		    );
		    sndref_num = SW_SNDREF_OHIT_SHIELDS;
		} 
		else
		{
		    s = (char *)malloc(
			STRLEN(tar_obj->name) + STRLEN(vect_name) +
			40 + (2 * 80)
		    );
		    sprintf(
			s,
"%s hit %s: %.2f damage to structure.",
			tar_obj->name, vect_name,
			dmg_structure
		    );
		    sndref_num = SW_SNDREF_OHIT_STRUCTURE;
		}
	    }
	    if((dmg_structure > 0.0f) || (dmg_shield > 0.0f))
		XSWPrint(s);
	    free(s);

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    &sound, sndref_num,
		    vol_left, vol_right,
		    0,
		    0 
		);
	}
	/* Someone else's weapon hit some object */
	else
	{
	    if(dmg_shield > 0.0f)
		sndref_num = SW_SNDREF_OHIT_SHIELDS;
	    else
		sndref_num = SW_SNDREF_OHIT_STRUCTURE;

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    &sound, sndref_num,
		    vol_left, vol_right,
		    0,
		    0
		);
	}
}

/*
 *	Object destroyed callback.
 */
void SWTDestroyNotify(
	const int src_obj_num,
	const int tar_obj_num,
	const int owner_obj,
	const int reason
)
{
	xsw_object_struct *src_obj, *tar_obj;
	xsw_net_parms_struct *net = &net_parms;

	/* Get pointers to source and target objects (may be NULL) */
	if(DBIsObjectGarbage(src_obj_num))   
	    src_obj = NULL;
	else
	    src_obj = xsw_object[src_obj_num];

	if(DBIsObjectGarbage(tar_obj_num))
	    tar_obj = NULL;
	else 
	    tar_obj = xsw_object[tar_obj_num];

	/* Check if the destroyed object is the player */
	if(tar_obj_num == net->player_obj_num)
	{
	    /* Player got destroyed */
	    char *s;

#if 0
	    /* Change viewscreen page to display the "destroyed"
	     * page, this will not change it immediatly but
	     * when we are disconnected (which is very soon)
	     * the "destroyed" page is the page that will be shown
	     */
	    VSChangePage(&bw->destroyed_page);
#endif

	    switch(reason)
	    {
	      case SW_DESTROY_REASON_SELF_DESTRUCT:
		s = STRDUP(
"Your vessel's power-core has over-loaded and detonated!"
		);
		break;
	      default:  /* SW_DESTROY_REASON_UNSPECIFIC */
		s = STRDUP(
"Your vessel has been destroyed!"
		);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}
	else
	{   
	    /* Some other object got destroyed */
	    char *s, *name;

	    if(tar_obj != NULL)
	    {
		name = STRDUP(tar_obj->name);
	    }
	    else
	    {
		name = (char *)malloc(80);
		sprintf(name, "Object #%i", tar_obj_num);
	    }
	    switch(reason)
	    {
	      case SW_DESTROY_REASON_SELF_DESTRUCT:
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"%s's power-core has over-loaded and detonated.",
		    name
		);
		break;
	      default:  /* SW_DESTROY_REASON_UNSPECIFIC */
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"%s has been destroyed.",
		    name 
		);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	    free(name);
	}
}

/*
 *      Manages curses.
 */
int SWTManageGUI(void *data)
{
	int i, events_handled, total_events_handled = 0;
	CDisplay *display;
	CEvent event;

	for(i = 0; i < total_curses_displays; i++)
	{
	    display = curses_display[i];
	    if(display == NULL)
		continue;

	    /* Reset events handled counter */
	    events_handled = 0;

	    /* Get next event if any */
	    if(CGetNextEvent(
		display,
		&event
	    ) <= 0)
		continue;

	    /* Begin calling event managers */
	    if(events_handled == 0)
		events_handled += BridgeManage(swt_bridge, &event);

	    total_events_handled += events_handled;
	}

	return(total_events_handled);
}

/*
 *	Manages sound.
 */
int SWTManageSound(void *data)
{
	return(0);
}

/*
 *	Manages network.
 */
int SWTManageNet(void *data)
{
	const time_t cur_sec = cur_systime;
	int total_events_handled = 0;
	xsw_net_parms_struct *net = &net_parms;

	/* Manage network */
	switch(net->connection_state)
	{
	  case XSW_CON_STATE_NOT_CONNECTED:
	    break;

	  case XSW_CON_STATE_NEGOTIATING:
	    /* Warn if logging in is taking longer than it should */
	    if(((net->con_start + 15l) <= cur_sec) &&
	       (net->login_delay_warning_last < 15l)
	    )
	    {
		XSWPrint(
"Waited 15 seconds for server response, still waiting..."
		);
		net->login_delay_warning_last = 15l;
	    }
	    else if(((net->con_start + 30l) <= cur_sec) &&
		    (net->login_delay_warning_last < 30l)
	    )
	    {
		XSWPrint(
"Waited 30 seconds for server response, still waiting..."
		);
		net->login_delay_warning_last = 30l;
	    }
	    else if(((net->con_start + 45l) <= cur_sec) &&
		    (net->login_delay_warning_last < 45l)
	    )
	    {
		XSWPrint(
"Waited 45 seconds (press F10 a few times to give up)..."
		);
		net->login_delay_warning_last = 45l;
	    }

	    /* Handle incoming data for login, version, or
	     * disconnect
	     */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net))
		break;

	    break;

	  case XSW_CON_STATE_CONNECTED:
	    /* Handle any incoming data */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net) ||
	       !XSW_CON_STATE_IS_CONNECTED(net)
	    )
		break;

	    /* Time to adjust network interval? */
	    if(auto_interval_tune.next <= cur_millitime)
	    {
		/* Adjust network interval */
		AutoIntervalTune();

		/* Schedual next network interval tune */
		auto_interval_tune.next = cur_millitime +
		    auto_interval_tune.interval;
	    }

	    /* Time to send out player object position to server? */
	    if(next.lplayer_pos_send <= cur_millitime)
	    {
		/* Get referance to player object number */
		int player_obj_num = net->player_obj_num;

		/* Pose object pose */
		NetSendObjPose(player_obj_num);

		/* Throttle position */
		NetSendObjThrottle(player_obj_num);

		/* Schedual next player position send */
		next.lplayer_pos_send = cur_millitime +
		    net->net_int;
	    }
	    break;

	  case XSW_CON_STATE_CLOSE_WAIT:
	    /* Handle incoming data for any subsequent data sent by
	     * the server and for recv() == 0
	     */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net))
		break;

	    break;
	}

	return(total_events_handled);
}


/* 
 *	Initializes SWTerm.
 */
int SWTInit(int argc, char *argv[])
{
	int i;
	const char *s, *arg, *home, *url = NULL;
	char cwd[PATH_MAX];
	CDisplay *dpy;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	/* Initialize time zone */
	tzset();

	/* Get the current working directory */
	if(getcwd(cwd, sizeof(cwd)) != NULL)
	    cwd[sizeof(cwd) - 1] = '\0';
	else
	    *cwd = '\0';

	/* Get the home directory */
	home = getenv("HOME");
	if(home == NULL)
	    home = "/";


	debug.level = DEBUG_LEVEL_NONE;
	debug.val = 0;

	cur_millitime = MilliTime();
	cur_systime = time(NULL);
	lapsed_millitime = 0;

	time_compensation = 1.0;
	
	next.update_check = cur_millitime;
	next.viewscreen = cur_millitime;
	next.bridge_std_redraw = cur_millitime;
	next.memory_clean = cur_millitime + XSW_MEMORY_CLEAN_INTERVAL;
	next.lplayer_pos_send = cur_millitime;
	next.spawn_particles = cur_millitime;
	next.wormhole_enter = cur_millitime;

	strncpy(dname.startup, cwd, sizeof(dname.startup));
	dname.startup[sizeof(dname.startup) - 1] = '\0';
	strncpy(dname.toplevel, DEF_XSW_TOPLEVEL_DIR, sizeof(dname.toplevel));
	dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';
	strncpy(dname.etc, DEF_XSW_ETC_DIR, sizeof(dname.etc));
	dname.etc[sizeof(dname.etc) - 1] = '\0'; 
	strncpy(dname.images, DEF_XSW_IMAGES_DIR, sizeof(dname.images));
	dname.images[sizeof(dname.images) - 1] = '\0';
	strncpy(dname.sounds, DEF_XSW_SOUNDS_DIR, sizeof(dname.sounds));
	dname.sounds[sizeof(dname.sounds) - 1] = '\0';
	strncpy(dname.downloads, home, sizeof(dname.downloads));
	dname.downloads[sizeof(dname.downloads) - 1] = '\0';

	xsw_font.std = NULL;
	xsw_font.std_bold = NULL;
	xsw_font.console_heading = NULL;
	xsw_font.console_standard = NULL;
	xsw_font.console_message = NULL;

	opt->cfg_version_major = PROG_VERSION_MAJOR;
	opt->cfg_version_minor = PROG_VERSION_MINOR;
	opt->cfg_version_release = PROG_VERSION_RELEASE;

	opt->cpu_priority = XSW_CPU_PRIORITY_MEDIUM;
	
	opt->units = XSW_UNITS_XSW;
	
	opt->log_client = False;
	opt->log_net = False;
	opt->log_errors = False;

	opt->debug_kb = False;

	opt->energy_saver_mode = False;
	opt->show_formal_label = XSW_SHOW_FORMAL_LABEL_AS_NEEDED;
	opt->show_fires = True;
	opt->show_smoke_trails = True;
	opt->max_particles = 500;
	opt->net_bandwidth = XSW_DEF_NET_BANDWIDTH;
	opt->net_object_outdated_time = XSW_DEF_OBJECT_OUTDATED_TIME;
	opt->show_net_errors = False;
	opt->show_server_errors = True;
	opt->local_updates = True;
	opt->sounds = XSW_SOUNDS_NONE;
	opt->music = False;

	opt->scanner_limiting = True;
	opt->notify_scanner_contacts = False;
	opt->save_on_exit = True;

	curses_display = NULL;
	total_curses_displays = 0;

	sound.server_type = SNDSERV_TYPE_NONE;
	sound.start_cmd = NULL;
	sound.con_arg = NULL;
	sound.con_data = NULL;
	sound.audio_mode_name = NULL;
	sound.bg_music_playid = NULL;
	sound.bg_music_sndref_num = -1;

	xsw_gctl = XSW_GCTL(calloc(1, sizeof(xsw_gctl_struct)));

	UNVLegendReset(&sw_legend);

	player_values.weapons_online = True;
	player_values.weapon_freq = 180.25;
	player_values.weapon_fire_heading = 0.0 * PI;
	player_values.weapon_yield = 1.0;
	player_values.auto_destruct_start_int = 20l;
	player_values.auto_destruct_last_sec_left = 0l;

	scanner_contact = NULL;
	total_scanner_contacts = 0;

	xsw_ocsn = NULL;
	total_ocsns = 0;

	xsw_imgref = NULL;
	total_imgrefs = 0;

	xsw_object = NULL;
	total_objects = 0;

	inrange_xsw_object = NULL;
	total_inrange_objects = 0;

	xsw_particle = NULL;
	total_particles = 0;

	xsw_sndref = NULL;
	total_sndrefs = 0;

	/* Reset player object to none (-1) */
	DBSetPlayerObject(-1);


	/* Set default fname */
	strncpy(
	    fname.cfg, PrefixPaths(cwd, XSW_DEF_CFG_FILE),
	    sizeof(fname.cfg)
	);
	fname.cfg[sizeof(fname.cfg) - 1] = '\0';

	strncpy(
	    fname.sndrefs, PrefixPaths(cwd, DEF_XSW_SOUND_SCHEME_FILE),
	    sizeof(fname.sndrefs)
	);
	fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';

	strncpy(
	    fname.ocsn, PrefixPaths(cwd, DEF_XSW_OCSN_FILE),
	    sizeof(fname.ocsn)
	);
	fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';

	strncpy(
	    fname.log, PrefixPaths(cwd, XSW_DEF_LOG_FILE),
	    sizeof(fname.log)
	);
	fname.log[sizeof(fname.log) - 1] = '\0';


	net->connection_state = XSW_CON_STATE_NOT_CONNECTED;
	net->socket = -1;
	net->flags = 0;
	net->player_obj_num = -1;
	net->player_obj_ptr = NULL;
	s = getenv("LOGNAME");
	net->login_name = (s != NULL) ?
	    STRDUP(s) :
	    STRDUP(XSW_DEF_GUEST_LOGIN_NAME);
	net->login_password = STRDUP(XSW_DEF_GUEST_LOGIN_PASSWORD);
	net->con_start = 0l;
	net->login_delay_warning_last = 0l;
	net->title = NULL;
	net->address = STRDUP(XSW_DEF_SERVER_ADDRESS);
	net->port = XSW_DEF_SERVER_PORT;
	net->net_int = XSW_SWSERV_DEF_UPDATE_INT;
	net->disconnect_send_count = 0;
	net->bad_send_count = 0;
	memset(net->recvbuf, 0x00, sizeof(net->recvbuf));
	net->recvbuf_len = 0;

	net_stat.rx_interval = 0l;
	net_stat.sx_interval = 0l;
	net_stat.rx_ilast = 0l;
	net_stat.sx_ilast = 0l;

	auto_interval_tune.state = True;
	auto_interval_tune.interval = AINT_DEF_TUNE_INT;
	auto_interval_tune.next = cur_millitime +
	    auto_interval_tune.interval;


	/* Initialize universe object management */
	UNVInit();

	/* Handle command line arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Data toplevel directory */
	    if(!strcasecmp(arg, "--data_dir") ||
	       !strcasecmp(arg, "--data-dir") ||
	       !strcasecmp(arg, "-data_dir") ||
	       !strcasecmp(arg, "-data-dir") ||
	       !strcasecmp(arg, "--d") ||
	       !strcasecmp(arg, "-d")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    strncpy(
			dname.toplevel, arg,
			sizeof(dname.toplevel)
		    );
		    dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';

		    strncpy(
			dname.etc,
			PrefixPaths(dname.toplevel, "etc"),
			sizeof(dname.etc)
		    );
		    dname.etc[sizeof(dname.etc) - 1] = '\0';

		    strncpy(
			dname.images,
			PrefixPaths(dname.toplevel, "images"),
			sizeof(dname.images)
		    );
		    dname.images[sizeof(dname.images) - 1] = '\0';

		    strncpy(
			dname.sounds,
			PrefixPaths(dname.toplevel, "sounds"),
			sizeof(dname.sounds)
		    );
		    dname.sounds[sizeof(dname.sounds) - 1] = '\0';
		}
		else
		{
		    fprintf(
			stderr,
"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Debug keyboard */
	    else if(!strcasecmp(arg, "--debug_keyboard") ||
		    !strcasecmp(arg, "-debug_keyboard") ||
		    !strcasecmp(arg, "--debug_kb") ||
		    !strcasecmp(arg, "-debug_kb") ||
		    !strcasecmp(arg, "--debugkb") ||
		    !strcasecmp(arg, "-debugkb")
	    )
	    {
		option.debug_kb = True;
	    }
	    /* URL? */
	    else if((*arg != '-') &&
		    (*arg != '+')
	    )
	    {
		url = arg;
	    }
	}


	/* Connect to curses library */
	total_curses_displays = 1;
	curses_display = (CDisplay **)malloc(
	    total_curses_displays * sizeof(CDisplay *)
	);
	if(curses_display == NULL)
	{
	    total_curses_displays = 0;
	    return(-1);
	}

	curses_display[0] = dpy = CWInit(argc, argv);
	if(dpy != NULL)
	{
	    /* Get colors */
#define GET_COLOR(_rc_,_bc_,_gc_)	\
 CWMatchColorCoeff(			\
  dpy,					\
  (_rc_), (_bc_), (_gc_),		\
  -1.0f, -1.0f, -1.0f			\
 )

	    swt_color.standard = GET_COLOR(
		1.0f, 1.0f, 1.0f
	    );
	    swt_color.warning = GET_COLOR(
		1.0f, 1.0f, 0.0f
	    );
	    swt_color.danger = GET_COLOR(
		1.0f, 0.0f, 0.0f
	    );

	    swt_color.title = GET_COLOR(
		1.0f, 1.0f, 1.0f
	    );
	    swt_color.messages = GET_COLOR(
		1.0f, 1.0f, 1.0f
	    );
	    swt_color.status = GET_COLOR(
		1.0f, 1.0f, 1.0f
	    );
	    swt_color.status_con = GET_COLOR(
		0.0f, 0.0f, 1.0f
	    );

	    swt_color.vs_obj_static = GET_COLOR(
		1.0f, 0.0f, 1.0f 
	    );
	    swt_color.vs_obj_dynamic = GET_COLOR(
		0.0f, 1.0f, 1.0f
	    );
	    swt_color.vs_obj_weapon = GET_COLOR(
		1.0f, 0.0f, 0.0f
	    );
	    swt_color.vs_obj_home = GET_COLOR(
		0.0f, 0.0f, 1.0f
	    );
	    swt_color.vs_obj_area = GET_COLOR(
		1.0f, 0.0f, 1.0f
	    );
	    swt_color.vs_obj_locked_on = GET_COLOR(
		1.0f, 1.0f, 0.0f
	    );

	    swt_color.vs_particle_fire = GET_COLOR(
		1.0f, 0.0f, 0.0f
	    );
	    swt_color.vs_particle_smoke = GET_COLOR(
		1.0f, 1.0f, 1.0f
	    );
	    swt_color.vs_particle_explosion = GET_COLOR(
		1.0f, 0.0f, 0.0f
	    );

#undef GET_COLOR
	}
	else
	{
	    /* Unable to initialize curses */
	    XSWPrintWarning(
"Unable to initialize curses."
	    );
	    return(-1);
	}

	/* Initialize the bridge window */
	swt_bridge = BridgeNew();
	if(swt_bridge == NULL)
	    return(-1);
	BridgeMap(swt_bridge);
	BridgeSetFocus(swt_bridge, 1);

	/* Initialize the reality engine */
	REngInit();


	/* Connect at startup? */
	if(url != NULL)
	{
	    SWTConnect(url);
	}
	else
	{
	    XSWPrint(PROG_NAME_FULL " ready, press F9 to connect.");
	}

#if 0
if(dpy != NULL)
{
 char s[256];
 sprintf(s, "standard=%i bold=%i",
  swt_color.standard, swt_color.bold
 );
 XSWPrint(s);
}
#endif

	return(0);
}

/*
 *	Manages one iteration of SWTerm.
 *
 *	Updates timmers and manages all resources.
 */
void SWTMainIteration(void *data)
{
	time_t	cur_ms, cur_sec,
		new_systime = time(NULL),
		new_millitime = MilliTime();
	xsw_net_parms_struct *net = &net_parms;

	/* Check if timmers have cycled */
	if((new_systime < cur_systime) ||
	   (new_millitime < cur_millitime)
	)
	{
	    /* Reset global timmers */
	    SWTResetTimmers(new_millitime, new_systime);
	    lapsed_millitime = 0l;
	    time_compensation = 1.0f;
	}
	else
	{
	    lapsed_millitime = new_millitime - cur_millitime;
	    time_compensation = (float)lapsed_millitime /
		(float)SW_CYCLE_LAPSE_MS;
	    if(time_compensation < 1.0f)
		time_compensation = 1.0f;
	}

	/* Now set global variable cur_millitime */
	cur_sec = cur_systime = new_systime;
	cur_ms = cur_millitime = new_millitime;

	/* Manage curses */
	SWTManageGUI(data);

	/* Manage sound */
	SWTManageSound(data);

	/* Manage network */
	SWTManageNet(data);

	/* Manage reality engine */
	REngManage(
	    net->player_obj_ptr,
	    cur_millitime, cur_systime, time_compensation
	);


	/* Redraw graphics */

	/* Redraw standard bridge graphics */
	if(next.bridge_std_redraw <= cur_millitime)
	{
	    /* Redraw bridge things that need to be redrawn often */
	    BridgeDrawSchedualed(
		swt_bridge,
		cur_millitime, cur_systime, time_compensation
	    );

	    /* Schedual next time to draw consoles */
	    next.bridge_std_redraw = cur_millitime +
		BRIDGE_STANDARD_REDRAW_INTERVAL;
	}

	/* ************************************************* */
	/* Clean/refresh memory */
	if(next.memory_clean <= cur_millitime)
	{
	    /* Clean, reclaim, memory and schedual next */
	    SWTReclaimGlobalMemory(False);

	    /* XSWReclaimGlobalMemory() already scheduals next */
	}
}

/*
 *	Manages SWTerm.
 *
 *	Pushes an SWTerm main loop and manages and blocks until
 *	SWTermMainPop() or SWTermMainPopAll() is called.
 */
void SWTMain(void *data)
{
	const int entering_runlevel = runlevel;
	runlevel++;
	while(runlevel > entering_runlevel)
	{
	    switch(option.cpu_priority)
	    {
	      case XSW_CPU_PRIORITY_HIGHEST:
		usleep(0);
		break;
	      case XSW_CPU_PRIORITY_HIGH:
		usleep(1000);
		break;
	      case XSW_CPU_PRIORITY_MEDIUM:
		usleep(SW_CYCLE_LAPSE_US / 2);
		break;
	      default:          /* XSW_CPU_PRIORITY_LOWEST */
		usleep(SW_CYCLE_LAPSE_US);
		break;
	    }
	     
	    SWTMainIteration(data);
	}
}

/*
 *	Pops off the highest SWTerm main loop.
 */
void SWTMainPop(void *data)
{
	if(runlevel > 0)
	    runlevel--;
}

/*
 *	Pops off all SWTerm main loops.
 */
void SWTMainPopAll(void *data)
{  
	runlevel = 0;
}

/*
 *	Shuts down SWTerm.
 *
 *	The configuration is not saved by this call, you should call
 *	CfgFileSave() prior to this call if you want to save the
 *	configuration.
 */
void SWTShutdown(void)
{
	int i;
	xsw_net_parms_struct *net = &net_parms;


	/* Disconnect */
	SWTDisconnect();
	free(net->login_name);
	net->login_name = NULL;
	free(net->login_password);
	net->login_password = NULL;
	free(net->title);
	net->title = NULL;


	/* Delete the bridge */
	BridgeDelete(swt_bridge);
	swt_bridge = NULL;


	/* Curses library resources */
	for(i = 0; i < total_curses_displays; i++)
	{
	    CWShutdown(curses_display[i]);
	}
	free(curses_display);
	curses_display = NULL;
	total_curses_displays = 0;


	/* Reality engine shutdown */
	REngShutdown();

	/* Shutdown sound server and free sound schemes */
	SoundShutdown(&sound);

	/* Shut down the game controller */
	free(xsw_gctl);
	xsw_gctl = NULL;

	/* Delete all the OCS Names */
	OCSNClear(&xsw_ocsn, &total_ocsns);

	/* Delete all scanner contacts */
	ScDeleteAll();

	/* Delete all objects */
	DBDeleteAllObjects();
	ParticleClear(&xsw_particle, &total_particles);

	/* Shutdown universe object management */
	UNVShutdown();
}

int main(int argc, char *argv[])
{
	int i, status;
	const char *arg;

	runlevel = 0;

	/* Set the signals to watch for */
	signal(SIGHUP, SWTSignalCB);
	signal(SIGINT, SWTSignalCB);
	signal(SIGQUIT, SWTSignalCB);
	signal(SIGABRT, SWTSignalCB);
	signal(SIGSEGV, SWTSignalCB);
	signal(SIGPIPE, SWTSignalCB);
	signal(SIGTERM, SWTSignalCB);
	signal(SIGCONT, SWTSignalCB);
	signal(SIGSTOP, SWTSignalCB);

	/* Handle help and version arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") ||
	       !strcasecmp(arg, "--h") ||
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?") ||
	       !strcasecmp(arg, "/?")
	    )
	    {
		XSWPrintHelpStdout();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		XSWPrintVersionStdout();
		return(0);
	    }
	}

	/* Initialize */
	status = SWTInit(argc, argv);
	if(status)
	{
	    SWTShutdown();
	    return(-status);
	}
	 
	/* Manage */
	SWTMain(NULL);

	/* Save the configuration to file? */
	if(option.save_on_exit)
	{
/*	    CfgFileSave(fname.cfg); */
/*	    UnivListFileSave(fname.universes); */
	}

	/* Shutdown */
	SWTShutdown();

	return(0);
}
