#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "prompt.h"


char *PromptGetS(CDisplay *display, Prompt *p);
int PromptGetI(CDisplay *display, Prompt *p);
long PromptGetL(CDisplay *display, Prompt *p);
double PromptGetF(CDisplay *display, Prompt *p);

void PromptSetLabel(CDisplay *display, Prompt *p, const char *label);
void PromptSetS(CDisplay *display, Prompt *p, const char *s);
void PromptSetI(CDisplay *display, Prompt *p, int i);
void PromptSetL(CDisplay *display, Prompt *p, long l);
void PromptSetF(CDisplay *display, Prompt *p, double d, int decimals);

int PromptInit(
	CDisplay *display,
	Prompt *p, Window parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int buf_len,
	const char *label,
	void (*activate_cb)(Prompt *, const char *, void *),
	void *activate_data
);
void PromptResize(
	CDisplay *display, Prompt *p,
	unsigned int width, unsigned int height
);
void PromptDraw(CDisplay *display, Prompt *p);
int PromptManage(CDisplay *display, Prompt *p, CEvent *event);
void PromptSetFocus(CDisplay *display, Prompt *p, char focus);
void PromptMap(CDisplay *display, Prompt *p);
void PromptUnmap(CDisplay *display, Prompt *p);
void PromptDestroy(CDisplay *display, Prompt *p);


/*
 *	Checks if control key was held down.
 */
#define PromptCtrl(c)	((c) & 037)


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)



/*
 *	Prompt value fetching.
 */
char *PromptGetS(CDisplay *display, Prompt *p)
{
	if((display == NULL) ||
	   (p == NULL)
	)
	    return(NULL);

	return(p->buf);
}

int PromptGetI(CDisplay *display, Prompt *p)
{
	if((display == NULL) ||
	   (p == NULL)
	)
	    return(0);

	return(atoi(p->buf));
}

long PromptGetL(CDisplay *display, Prompt *p)
{
	if((display == NULL) ||
	   (p == NULL)
	)
	    return(0);

	return(atol(p->buf));
}

double PromptGetF(CDisplay *display, Prompt *p)
{
	if((display == NULL) ||
	   (p == NULL)
	)
	    return(0);

	return(atof(p->buf));
}


/*
 *	Prompt value setting.
 */
void PromptSetLabel(CDisplay *display, Prompt *p, const char *label)
{
	if((display == NULL) || (p == NULL))
	    return;

	free(p->label);
	p->label = STRDUP(label);

	/* Need to recalculate buf_start_pos */
	if(p->label == NULL)
	    p->buf_start_pos = 0;
	else
	    p->buf_start_pos = strlen(p->label) + 1;
}

void PromptSetS(CDisplay *display, Prompt *p, const char *s)
{
	if((display == NULL) || (p == NULL))
	    return;

	if(p->buf == NULL)
	    return;

	strncpy(p->buf, (s != NULL) ? s : "", p->buf_len);
	p->buf[p->buf_len] = '\0';

	/* Move cursor position to end */
	p->buf_cursor_pos = strlen(p->buf);

	/* Redraw as needed */
	if(p->map_state)
	    PromptDraw(display, p);
}

void PromptSetI(CDisplay *display, Prompt *p, int i)
{
	char num_str[40];

	if((display == NULL) || (p == NULL))
	    return;

	if(p->buf == NULL)   
	    return;

	sprintf(num_str, "%i", i);
	PromptSetS(display, p, num_str);
}

void PromptSetL(CDisplay *display, Prompt *p, long l)
{
	char num_str[128];

	
	if((display == NULL) ||
	   (p == NULL)
	)
	    return;

	if(p->buf == NULL)
	    return;

	sprintf(num_str, "%ld", l);
	PromptSetS(display, p, num_str);
}

void PromptSetF(CDisplay *display, Prompt *p, double d, int decimals)
{
	char num_str[80];
	char num_str2[128];

	if((display == NULL) || (p == NULL))
	    return;

	if(p->buf == NULL)
	    return;

	sprintf(num_str, "%%.%ilf", decimals);
	sprintf(num_str2, num_str, d);
	PromptSetS(display, p, num_str2);
}


/*
 *	Initializes the prompt.
 */
int PromptInit(
	CDisplay *display,
	Prompt *p, Window parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int buf_len,
	const char *label,
	void (*activate_cb)(Prompt *, const char *, void *),
	void *activate_data
)
{
	int i;

	if((display == NULL) || (parent == None) || (p == NULL))
	    return(-1);

	memset(p, 0x00, sizeof(Prompt));

	p->map_state = False;
	p->is_in_focus = False;
	p->x = x;
	p->y = y;
	p->width = width;
	p->height = height;
	p->w = parent;

	p->color_num = CWMatchColorCoeff(
	    display,
	    1.0f, 1.0f, 1.0f,
	    0.0f, 0.0f, 0.0f
	);

	p->nhistory_items = 40;
	p->history = (char **)malloc(p->nhistory_items * sizeof(char *));
	for(i = 0; i < p->nhistory_items; i++)
	    p->history[i] = NULL;
	p->cur_history_pos = -1;

	p->activate_cb = activate_cb;
	p->activate_data = activate_data;


	/* Allocate buffer (add one extra char) */
	if(buf_len < 1)
	    buf_len = 1;
	p->buf = (char *)malloc((buf_len + 1) * sizeof(char));
	if(p->buf == NULL)
	{
	    p->buf_len = 0;
	    return(-1);
	}

	p->buf_len = buf_len;
	*p->buf = '\0';			/* Set first char null */
	p->buf[p->buf_len] = '\0';	/* Set last char null */

	p->buf_scroll_pos = 0;
	p->buf_cursor_pos = 0;

	/* Set label */
	p->label = NULL;
	PromptSetLabel(display, p, label);

	return(0);
}

void PromptResize(
	CDisplay *display, Prompt *p,
	unsigned int width, unsigned int height
)
{
	if((display == NULL) || (p == NULL))
	    return;

	if(width < 1)
	    width = 1;
	if(height < 1)
	    height = 1;

	p->width = width;
	p->height = height;
}

void PromptDraw(CDisplay *display, Prompt *p)
{
	int x, y, x2, len;
	int scroll_pos, scroll_end_pos, cursor_pos;
	Window w;
	char tmp_char;

	if((display == NULL) || (p == NULL))
	    return;

	/* Do not draw if not mapped */
	if(!p->map_state)
	    return;

	/* Get window to draw to */
	w = p->w;

	CWSetColor(display, w, p->color_num, CColorFlagBold | CColorFlagReverse);

	/* Clear entire prompt area */
	for(x = p->x, y = p->y, x2 = p->x + (int)p->width;
	    x < x2;
	    x++
	)
	    CWDrawCharacter(
		display,
		w,
		x, y,
		' '
	    );


	/* Get starting values */
	x = p->x;
	y = p->y;

	/* Draw label (if prompt has one) and add one extra space */
	if(p->label != NULL)
	{
	    len = strlen(p->label);

	    CWDrawStringPosition(
		display,
		w,
		x, y,
		p->label
	    );
	    x += len;

	    /* Draw a space to padd between label and text value */
	    CWDrawStringPosition(
		display,
		w,       
		x, y,   
		" "
	    ); 
	    x += 1;
	}


	/* Get sanitized scrolled buffer position */
	scroll_pos = p->buf_scroll_pos;
	if(scroll_pos >= p->buf_len)
	    scroll_pos = p->buf_len - 1;
	if(scroll_pos < 0)
	    scroll_pos = 0;

	/* Get sanitized cursor position */
	cursor_pos = p->buf_cursor_pos;
	if(cursor_pos >= p->buf_len)
	    cursor_pos = p->buf_len - 1;
	if(cursor_pos < 0)
	    cursor_pos = 0;

	/* Draw text value if exists */
	if((p->buf != NULL) && (p->buf_len > 0))
	{
	    /* Calculate length of characters visable */
	    len = (int)p->width - p->buf_start_pos - 1;
	    if(len < 0)
		len = 0;

	    /* Calculate scroll end position (can be at buffer end) */
	    scroll_end_pos = scroll_pos + len;
	    if(scroll_end_pos > p->buf_len)
		scroll_end_pos = p->buf_len;
	    if(scroll_end_pos < 0)
		scroll_end_pos = 0;

	    /* Playing a trick here, store character at
	     * scroll_end_pos then set it to null.
	     * After redrawing, set it back.
	     */
	    tmp_char = p->buf[scroll_end_pos];
	    p->buf[scroll_end_pos] = '\0';

	    /* Draw text value */
	    CWDrawStringPosition(
		display,
		w,
		x, y,
		&p->buf[scroll_pos]
	    );
	    x += len;


	    /* Set back tempory character at scrolled end position */
	    p->buf[scroll_end_pos] = tmp_char;


	    /* Draw a space to padd after text value */
	    CWDrawStringPosition(
		display,
		w,
		x, y,
		" "
	    );
	    x += 1;


	    /* Make the character the cursor is on `stand out' */
	    tmp_char = p->buf[cursor_pos];
	    if((tmp_char == '\0') ||	/* Use '_' so standout shows */
	       (tmp_char == ' ')
	    )
		tmp_char = '_';

	    x2 = p->x + p->buf_start_pos +
		(cursor_pos - scroll_pos);

	    standout();
	    CWDrawCharacter(
		display,
		w,
		x2, y,
		tmp_char
	    );
	    standend();
	     

	}

	/* Move cursor back to the starting position of prompt */
	CWMoveCursor(
	    display,
	    w,
	    p->x, p->y
	);
}

int PromptManage(CDisplay *display, Prompt *p, CEvent *event)
{
	int i, c, buf_cursor_pos, chars_visable, text_value_len;
	int events_handled = 0;

	if((display == NULL) || (p == NULL) || (event == NULL))
	    return(events_handled);

	switch(event->type)
	{
	  /* **************************************************** */
	  case CKeyPress:
	    if(!p->is_in_focus || !p->map_state)
		break;

	    if((p->buf == NULL) || (p->buf_len <= 0))
		break;

	    /* Get key character value */
	    c = event->ckey.key;

	    /* Get position of cursor in text value buffer */
	    buf_cursor_pos = p->buf_cursor_pos;

	    if(buf_cursor_pos >= p->buf_len)
		buf_cursor_pos = p->buf_len - 1;
	    if(buf_cursor_pos < 0)
		buf_cursor_pos = 0;

	    /* Calculate characters visable */
	    chars_visable = (int)p->width - p->buf_start_pos - 1;
	    if(chars_visable < 0)
		chars_visable = 0;


	    /* Check what character we got */
	    switch(c)
	    {
	      case KEY_LEFT:
		buf_cursor_pos--;
		break;

	      case KEY_RIGHT:
		buf_cursor_pos++;
		break;

	      /* Home */
	      case 0x00000001:
		buf_cursor_pos = 0;
		break;

	      /* End */
	      case 0x00000005:
		buf_cursor_pos = STRLEN(p->buf);
		break;

	      case KEY_UP:
		p->cur_history_pos = CLIP(
		    (p->cur_history_pos + 1),
		    0, p->nhistory_items - 1
		);
		if((p->cur_history_pos >= 0) && (p->cur_history_pos < p->nhistory_items))
		{
		    const char *history_item = p->history[p->cur_history_pos];
		    if(history_item != NULL)
			strncpy(p->buf, history_item, p->buf_len);
		    else
			p->buf[0] = '\0';
		    buf_cursor_pos = STRLEN(p->buf);
		}
		break;

	      case KEY_DOWN:
		p->cur_history_pos = CLIP(
		    (p->cur_history_pos - 1),
		    0, p->nhistory_items - 1
		);
		if((p->cur_history_pos >= 0) && (p->cur_history_pos < p->nhistory_items))
		{
		    const char *history_item = p->history[p->cur_history_pos];
		    if(history_item != NULL)
			strncpy(p->buf, history_item, p->buf_len);
		    else
			p->buf[0] = '\0';
		    buf_cursor_pos = STRLEN(p->buf);
		}
		break;

	      case KEY_ENTER:	/* Enter */
	      case KEY_NUMPADENTER:
		if(p->activate_cb != NULL)
		{
		    /* Record history? */
		    if(p->nhistory_items > 0)
		    {
			int i;
			for(i = p->nhistory_items - 1; i > 0; i--)
			{
			    free(p->history[i]);
			    p->history[i] = STRDUP(p->history[i - 1]);
			}
			i = 0;
			free(p->history[i]);
			p->history[i] = STRDUP(p->buf);
			p->cur_history_pos = -1;
		    }
		    p->activate_cb(
			p,
			p->buf,
			p->activate_data
		    );
		    p->buf[0] = '\0';
		    buf_cursor_pos = 0;
		}
		break;

	      case PromptCtrl('l'):	/* Redraw */
		break;

	      case KEY_BACKSPACE:	/* Backspace */
#if (KEY_BACKSPACE != 0x00000008)
	      case 0x00000008:
#endif
		if(buf_cursor_pos > 0)
		{
		    buf_cursor_pos--;

		    for(i = buf_cursor_pos; i < p->buf_len; i++)
		    {
			p->buf[i] = p->buf[i + 1];
			if(p->buf[i] == '\0')
			    break;
		    }
		}
		break;

	      case KEY_DC:	/* Delete */
	      case PromptCtrl('d'):
	      case KEY_DELETE:
		for(i = buf_cursor_pos; i < p->buf_len; i++)
		{
		    p->buf[i] = p->buf[i + 1];
		    if(p->buf[i] == '\0')
			break;
		}
		break;

	      case PromptCtrl('k'):	/* Clear line */
		p->buf[0] = '\0';
		buf_cursor_pos = 0;
	        break;

	      default:
		/* Shift characters */
		for(i = p->buf_len - 1; i > buf_cursor_pos; i--)
		    p->buf[i] = p->buf[i - 1];

		/* Null terminate, this may be needed! */
		p->buf[p->buf_len] = '\0';

		/* Set new character */
		p->buf[buf_cursor_pos] = (char)c;

		/* Increment cursor position */
		buf_cursor_pos++;

		break;
	    }

	    /* Get length of text value buffer contents */
	    text_value_len = strlen(p->buf);

	    /* Sanitize new cursor position in buffer */
	    if(buf_cursor_pos > text_value_len)
		buf_cursor_pos = text_value_len;
	    if(buf_cursor_pos >= p->buf_len)
		buf_cursor_pos = p->buf_len - 1;
	    if(buf_cursor_pos < 0)
		buf_cursor_pos = 0;

	    /* Adjust scrolled position as needed */
	    if(buf_cursor_pos < p->buf_scroll_pos)
	    {
		p->buf_scroll_pos = buf_cursor_pos;
	    }
	    else if(buf_cursor_pos >= (p->buf_scroll_pos + chars_visable))
	    {
	        p->buf_scroll_pos = buf_cursor_pos - chars_visable + 1;
	    }
	    if(p->buf_scroll_pos >= p->buf_len)
		p->buf_scroll_pos = p->buf_len - 1;
	    if(p->buf_scroll_pos < 0)
		p->buf_scroll_pos = 0;

	    /* Set new cursor position */
	    p->buf_cursor_pos = buf_cursor_pos;

	    PromptDraw(display, p);

	    events_handled++;

	    break;

	  /* **************************************************** */
	  case CKeyRelease:
	    if(!p->is_in_focus ||
	       !p->map_state
	    )
		break;

	    if((p->buf == NULL) || (p->buf_len < 1))
		break;


	    break;
	}



	return(events_handled);
}

void PromptSetFocus(CDisplay *display, Prompt *p, char focus)
{
	if((display == NULL) || (p == NULL))
	    return;

	p->is_in_focus = focus;
}

void PromptMap(CDisplay *display, Prompt *p)
{
	if((display == NULL) || (p == NULL))
	    return;

	p->cur_history_pos = -1;
	p->map_state = True;
	PromptDraw(display, p);
}

void PromptUnmap(CDisplay *display, Prompt *p)
{
	if((display == NULL) || (p == NULL))
	    return;

	/* If was not mapped, then clear area */
	if(p->map_state)
	{
	    CWClearArea(
		display,
		p->w,
		p->x, p->y,
		p->width, p->height
	    );

/*
	    CWSync(display, 0);
 */
	}

	/* Prompt should go out of focus when unmapped */
	PromptSetFocus(display, p, 0);

	p->map_state = False;
}

void PromptDestroy(CDisplay *display, Prompt *p)
{
	int i;

	if((display == NULL) || (p == NULL))
	    return;

	free(p->buf);
	free(p->label);

	for(i = 0; i < p->nhistory_items; i++)
	    free(p->history[i]);
	free(p->history);

	memset(p, 0x00, sizeof(Prompt));
}
