#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/swsoundcodes.h"
#include "../include/unvmath.h"

#include "sound.h"
#include "xsw.h"


static int SC_GET_OBJECT_NUM(xsw_object_struct *obj);
int ScIsAllocated(int entry_num);

int ScAddObjectInContact(xsw_object_struct *obj);
void ScRemoveObjectFromContact(xsw_object_struct *obj);

int ScIsObjectInContact(xsw_object_struct *obj);

void ScDelete(int entry_num);
void ScDeleteAll(void);
void ScReclaim(void);

int ScHandleContacts(int object_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the object number of the specified object pointer.
 *
 *	Important: If the object is garbage, the valid object number
 *	will still be returned.
 */
static int SC_GET_OBJECT_NUM(xsw_object_struct *obj)
{
	int i;
	xsw_object_struct **list = xsw_object;

	if(obj == NULL)
	    return(-1);

	for(i = 0; i < total_objects; i++)
	{
	    if(list[i] == obj)
		return(i);
	}

	return(-1);
}

/*
 *	Checks if the scanner contact entry is allocated.
 */
int ScIsAllocated(int entry_num)
{
	if((entry_num >= total_scanner_contacts) ||
	   (entry_num < 0) ||
	   (scanner_contact == NULL)
	)
	    return(0);
	else if(scanner_contact[entry_num] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Adds an object to the scanner contact list, allocating
 *	more scanner contact entries as needed.
 *
 *	Warning, does not check if object is already in contact.
 */
int ScAddObjectInContact(xsw_object_struct *obj)
{
	int i, n;
	xsw_scanner_contact_struct **list, *sc;

	if(obj == NULL)
	    return(-1);

	/* Sanitize total */
	if(total_scanner_contacts < 0)
	    total_scanner_contacts = 0;

	/* Check for available allocated entry */
	list = scanner_contact;
	for(i = 0; i < total_scanner_contacts; i++)
	{
	    sc = list[i];
	    if(sc == NULL)
		break;

	    if(DBIsObjectGarbage(sc->obj_num))
		break;
	}
	if(i < total_scanner_contacts)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_scanner_contacts;
	    total_scanner_contacts = n + 1;

	    /* Reallocate pointers */
	    scanner_contact = (xsw_scanner_contact_struct **)realloc(
		scanner_contact,
		total_scanner_contacts * sizeof(xsw_scanner_contact_struct *)
	    );
	    if(scanner_contact == NULL)
	    {
		total_scanner_contacts = 0;
		return(-1);
	    }

	    /* Reset newly allocated pointer so it gets allocated */
	    scanner_contact[n] = NULL;
	}

	/* Allocate a new scanner contact */
	if(scanner_contact[n] == NULL)
	{
	    scanner_contact[n] = (xsw_scanner_contact_struct *)calloc(
	        1, sizeof(xsw_scanner_contact_struct)
	    );
	    if(scanner_contact[n] == NULL)
	        return(-1);
	}

	sc = scanner_contact[n];

	/* Set values */
	sc->obj_num = SC_GET_OBJECT_NUM(obj);

	return(n);
}

/*
 *	Removes an object from the scanner contacts list.
 */
void ScRemoveObjectFromContact(xsw_object_struct *obj)
{
	/* Is the object in the scanner contacts list? */
	const int i = ScIsObjectInContact(obj);

	/* If it was in the contacts list then remove it */
	if(ScIsAllocated(i))
	    scanner_contact[i]->obj_num = -1;
}


/*
 *	Checks if object_num is in scanner contacts list.
 *
 *	If it is, then the scanner contact index number is returned
 *	or -1 if the object is not found.
 */
int ScIsObjectInContact(xsw_object_struct *obj)
{
	int i, obj_num;
	xsw_scanner_contact_struct **list;

	if(obj == NULL)
	    return(-1);

	obj_num = SC_GET_OBJECT_NUM(obj);
	if(obj_num < 0)
	    return(-1);

	list = scanner_contact;
	for(i = 0; i < total_scanner_contacts; i++)
	{
	    if(list[i] == NULL)
		continue;

	    if(list[i]->obj_num == obj_num)
		return(i);
	}

	return(-1);
}


/*
 *	Deletes a scanner contact entry.
 */
void ScDelete(int entry_num)
{
	if(!ScIsAllocated(entry_num))
	    return;

	free(scanner_contact[entry_num]);
	scanner_contact[entry_num] = NULL;
}

/*
 *	Deletes all scanner contact structures.
 */
void ScDeleteAll()
{
	int i;

	for(i = 0; i < total_scanner_contacts; i++)
	    ScDelete(i);

	free(scanner_contact);
	scanner_contact = NULL;
	total_scanner_contacts = 0;
}

/*
 *	Deletes unused scanner contacts.
 */
void ScReclaim(void)
{
	int i, h;
	xsw_scanner_contact_struct **ptr;

	/* Get highest entry in use */
	for(i = 0, h = -1, ptr = scanner_contact;
	    i < total_scanner_contacts;
	    i++, ptr++
	)
	{
	    if(*ptr == NULL)
		continue;

	    /* Object must be valid */
	    if(DBIsObjectGarbage((*ptr)->obj_num))
		continue;

	    h = i;
	}

	/* Free unused entries */
	for(i = h + 1;
	    i < total_scanner_contacts;
	    i++
	)
	    ScDelete(i);

	/* Adjust global variable total_scanner_contacts */
	total_scanner_contacts = h + 1;

	/* Reallocate pointers */
	if(total_scanner_contacts > 0)
	{
	    scanner_contact = (xsw_scanner_contact_struct **)realloc(
		scanner_contact,
		total_scanner_contacts * sizeof(xsw_scanner_contact_struct *)
	    );
	    if(scanner_contact == NULL)
	    {
		total_scanner_contacts = 0;
		return;
	    }
	}
	else
	{
	    free(scanner_contact);
	    scanner_contact = NULL;
	    total_scanner_contacts = 0;
	}
}


/*
 *	Manages scanner contacts.
 *
 *	Calls to this function should be timmed properly, this
 *	function does NOT need to be called every time per loop.
 *
 *	This function is only used on the client to notify
 *	when a new object comes into scanner range of object_num.
 */
int ScHandleContacts(int object_num)
{
	bool_t played_scanner_beep = False;
	int i, status;
	float scanner_range;
	xsw_sound_struct *snd = &sound;
	xsw_object_struct *player_obj, *tar_obj, **list;

	/* Skip if global option specifies not to do scanner notifying */
	if(!option.notify_scanner_contacts)
	    return(0);

	/* Check if object_num is valid */
	if(DBIsObjectGarbage(object_num))
	    return(-1);
	else
	    player_obj = xsw_object[object_num];

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(player_obj);
	if(scanner_range <= 0.0f)
	    return(-1);

	/* Begin updating the contacts list */
	list = inrange_xsw_object;
	for(i = 0; i < total_inrange_objects; i++)
	{
	    tar_obj = list[i];
	    if(tar_obj == NULL)
		continue;
	    if(tar_obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Skip the following object types plus outdated ones and
	     * the player itself
	     */
	    if((tar_obj == player_obj) ||
	       (tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
/* Animated objects are usually explosions, but sometimes not, so
 * skip these for now even though some may not be explosions
 */
	       (tar_obj->type == SW_OBJ_TYPE_ANIMATED)
	    )
		continue;

	    /* Is this object already in the scanner contacts list? */
	    if(ScIsObjectInContact(tar_obj) > -1)
	    {
		/* The object is in the scanner contact list, check if
		 * it is within scanner range
		 */

	        /* Outdated? */
		if((tar_obj->last_updated + option.net_object_outdated_time)
		    < cur_millitime
		)
		{
		    status = 0;
		}
		else
		{
		    status = DBIsObjectInEffectiveScannerRangePtr(
			player_obj, tar_obj, scanner_range
		    );
		}
		/* Is this object not in scanner range JUST NOW? */
		if(!status)
		{
		    /* The object JUST LEFT scanner range */

		    /* Print message */
		    char *s = (char *)malloc(
			STRLEN(tar_obj->name) + 40 + 80
		    );
		    if((tar_obj->sect_x == player_obj->sect_x) &&
		       (tar_obj->sect_y == player_obj->sect_y) &&
		       (tar_obj->sect_z == player_obj->sect_z)
		    )
		    {
		        sprintf(
			    s,
"%s just left scanner range, bearing %.0f'.",
			    tar_obj->name,
			    RADTODEG(
				MuCoordinateDeltaVector(
				    tar_obj->x - player_obj->x,
				    tar_obj->y - player_obj->y
				)
			    )
			);
		    }
		    else
		    {
			sprintf(
			    s,
"%s just left scanner range, bearing %.0f'.",
			    tar_obj->name,
			    RADTODEG(
				MuCoordinateDeltaVector(
				    tar_obj->sect_x - player_obj->sect_x,
				    tar_obj->sect_y - player_obj->sect_y
				)
			    )
			);
		    }
		    XSWPrint(s);
		    free(s);

		    /* Remove object from this scanner contact entry */
		    ScRemoveObjectFromContact(tar_obj);
		}
	    }
	    else
	    {
		/* The object was not in the scanner contacts list,
		 * check if the object is within scanner range
		 */

		/* Outdated? */ 
		if((tar_obj->last_updated + option.net_object_outdated_time)
		    < cur_millitime
		)
		{
		    status = 0;
		}
		else     
		{
		    status = DBIsObjectInEffectiveScannerRangePtr(
			player_obj, tar_obj, scanner_range
		    );
		}
		/* Is the object in scanner range now? */
		if(status)
		{
		    /* The object just entered scanner range */

		    /* Print message */
		    char *s = (char *)malloc(
			STRLEN(tar_obj->name) + 40 + 80
		    );
		    sprintf(
			s,
"%s just entered scanner range, bearing %.0f'.",
			tar_obj->name,
			RADTODEG(
			    MuCoordinateDeltaVector(
				tar_obj->x - player_obj->x,
				tar_obj->y - player_obj->y
			    )
			)
		    );
		    XSWPrintBold(s);
		    free(s);

		    /* Add object to scanner contacts */
		    ScAddObjectInContact(tar_obj);

		    /* Play scanner contacts sound */
		    if((option.sounds >= XSW_SOUNDS_EVENTS) &&
		       !played_scanner_beep
		    )
		    {
			SoundPlay(
			    snd,
			    SW_SNDREF_CONTACTS_BEEP,
			    1.0f, 1.0f,
			    0,
			    1
			);
			played_scanner_beep = True;
		    }
		}
	    }
	}

	/* Reclaim scanner contacts memory */
	ScReclaim();

	return(0);
}
